# This file is part of cloud-init. See LICENSE file for license information.
import logging
import os
from io import BytesIO
from unittest import mock

import configobj
import pytest

from cloudinit import util
from cloudinit.config import cc_mcollective
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import skipUnlessJsonSchema
from tests.unittests.util import get_cloud

LOG = logging.getLogger(__name__)


STOCK_CONFIG = """\
main_collective = mcollective
collectives = mcollective
libdir = /usr/share/mcollective/plugins
logfile = /var/log/mcollective.log
loglevel = info
daemonize = 1

# Plugins
securityprovider = psk
plugin.psk = unset

connector = activemq
plugin.activemq.pool.size = 1
plugin.activemq.pool.1.host = stomp1
plugin.activemq.pool.1.port = 61613
plugin.activemq.pool.1.user = mcollective
plugin.activemq.pool.1.password = marionette

# Facts
factsource = yaml
plugin.yaml = /etc/mcollective/facts.yaml
"""


@pytest.fixture
def server_cfg(tmp_path):
    return str(tmp_path / cc_mcollective.SERVER_CFG)


@pytest.fixture
def pricert_file(tmp_path):
    return str(tmp_path / cc_mcollective.PRICERT_FILE)


@pytest.fixture
def pubcert_file(tmp_path):
    return str(tmp_path / cc_mcollective.PUBCERT_FILE)


@pytest.mark.usefixtures("fake_filesystem")
class TestConfig:
    def test_basic_config(self):
        cfg = {
            "mcollective": {
                "conf": {
                    "loglevel": "debug",
                    "connector": "rabbitmq",
                    "logfile": "/var/log/mcollective.log",
                    "ttl": "4294957",
                    "collectives": "mcollective",
                    "main_collective": "mcollective",
                    "securityprovider": "psk",
                    "daemonize": "1",
                    "factsource": "yaml",
                    "direct_addressing": "1",
                    "plugin.psk": "unset",
                    "libdir": "/usr/share/mcollective/plugins",
                    "identity": "1",
                },
            },
        }
        expected = cfg["mcollective"]["conf"]

        cc_mcollective.configure(cfg["mcollective"]["conf"])
        contents = util.load_binary_file(cc_mcollective.SERVER_CFG)
        contents = configobj.ConfigObj(BytesIO(contents))
        assert expected == dict(contents)

    def test_existing_config_is_saved(self, server_cfg):
        cfg = {"loglevel": "warn"}
        util.write_file(server_cfg, STOCK_CONFIG)
        cc_mcollective.configure(config=cfg, server_cfg=server_cfg)
        assert os.path.exists(server_cfg)
        assert os.path.exists(server_cfg + ".old")
        assert util.load_text_file(server_cfg + ".old") == STOCK_CONFIG

    def test_existing_updated(self, server_cfg):
        cfg = {"loglevel": "warn"}
        util.write_file(server_cfg, STOCK_CONFIG)
        cc_mcollective.configure(config=cfg, server_cfg=server_cfg)
        cfgobj = configobj.ConfigObj(server_cfg)
        assert cfg["loglevel"] == cfgobj["loglevel"]

    def test_certificats_written(self, pricert_file, pubcert_file, server_cfg):
        # check public-cert and private-cert keys in config get written
        cfg = {
            "loglevel": "debug",
            "public-cert": "this is my public-certificate",
            "private-cert": "secret private certificate",
        }

        cc_mcollective.configure(
            config=cfg,
            server_cfg=server_cfg,
            pricert_file=pricert_file,
            pubcert_file=pubcert_file,
        )

        found = configobj.ConfigObj(server_cfg)

        # make sure these didnt get written in
        assert "public-cert" not in found
        assert "private-cert" not in found

        # these need updating to the specified paths
        assert found["plugin.ssl_server_public"] == pubcert_file
        assert found["plugin.ssl_server_private"] == pricert_file

        # and the security provider should be ssl
        assert found["securityprovider"] == "ssl"

        assert util.load_text_file(pricert_file) == cfg["private-cert"]
        assert util.load_text_file(pubcert_file) == cfg["public-cert"]


class TestHandler:
    @mock.patch("cloudinit.config.cc_mcollective.subp")
    @mock.patch("cloudinit.config.cc_mcollective.util")
    def test_mcollective_install(self, mock_util, mock_subp):
        cc = get_cloud()
        cc.distro = mock.MagicMock()
        mock_util.load_binary_file.return_value = b""
        mycfg = {"mcollective": {"conf": {"loglevel": "debug"}}}
        cc_mcollective.handle("cc_mcollective", mycfg, cc, [])
        assert cc.distro.install_packages.called
        install_pkg = cc.distro.install_packages.call_args_list[0][0][0]
        assert install_pkg == ["mcollective"]

        assert mock_subp.subp.called
        assert mock_subp.subp.call_args_list[0][0][0] == [
            "service",
            "mcollective",
            "restart",
        ]


class TestMcollectiveSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        [
            # Disallow undocumented keys client 'mcollective' without error
            (
                {"mcollective": {"customkey": True}},
                "mcollective: Additional properties are not allowed",
            ),
            # Allow undocumented keys client keys below 'conf' without error
            ({"mcollective": {"conf": {"customkey": 1}}}, None),
            # Don't allow undocumented keys that don't match expected type
            (
                {"mcollective": {"conf": {"": {"test": None}}}},
                "does not match any of the regexes:",
            ),
            (
                {"mcollective": {"conf": {"public-cert": 1}}},
                "mcollective.conf.public-cert: 1 is not of type 'string'",
            ),
        ],
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        if error_msg is None:
            validate_cloudconfig_schema(config, get_schema(), strict=True)
        else:
            with pytest.raises(SchemaValidationError, match=error_msg):
                validate_cloudconfig_schema(config, get_schema(), strict=True)



# This file is part of cloud-init. See LICENSE file for license information.

import re

from cloudinit.distros.parsers.sys_conf import SysConf

# Lots of good examples @
# http://content.hccfl.edu/pollock/AUnix1/SysconfigFilesDesc.txt


class TestSysConfHelper:

    def test_parse_no_change(self):
        contents = """# A comment
USESMBAUTH=no
KEYTABLE=/usr/lib/kbd/keytables/us.map
SHORTDATE=$(date +%y:%m:%d:%H:%M)
HOSTNAME=blahblah
NETMASK0=255.255.255.0
# Inline comment
LIST=$LOGROOT/incremental-list
IPV6TO4_ROUTING='eth0-:0004::1/64 eth1-:0005::1/64'
ETHTOOL_OPTS="-K ${DEVICE} tso on; -G ${DEVICE} rx 256 tx 256"
USEMD5=no"""
        conf = SysConf(contents.splitlines())
        assert conf["HOSTNAME"] == "blahblah"
        assert conf["SHORTDATE"] == "$(date +%y:%m:%d:%H:%M)"
        # Should be unquoted
        assert (
            conf["ETHTOOL_OPTS"]
            == "-K ${DEVICE} tso on; -G ${DEVICE} rx 256 tx 256"
        )
        assert contents == str(conf)

    def test_parse_shell_vars(self):
        contents = "USESMBAUTH=$XYZ"
        conf = SysConf(contents.splitlines())
        assert contents == str(conf)
        conf = SysConf("")
        conf["B"] = "${ZZ}d apples"
        # Should be quoted
        assert 'B="${ZZ}d apples"' == str(conf)
        conf = SysConf("")
        conf["B"] = "$? d apples"
        assert 'B="$? d apples"' == str(conf)
        contents = 'IPMI_WATCHDOG_OPTIONS="timeout=60"'
        conf = SysConf(contents.splitlines())
        assert "IPMI_WATCHDOG_OPTIONS=timeout=60" == str(conf)

    def test_parse_adjust(self):
        contents = 'IPV6TO4_ROUTING="eth0-:0004::1/64 eth1-:0005::1/64"'
        conf = SysConf(contents.splitlines())
        # Should be unquoted
        assert "eth0-:0004::1/64 eth1-:0005::1/64" == conf["IPV6TO4_ROUTING"]
        conf["IPV6TO4_ROUTING"] = "blah \tblah"
        contents2 = str(conf).strip()
        # Should be requoted due to whitespace
        assert re.search(r"IPV6TO4_ROUTING='blah\s+blah'", contents2)

    def test_parse_no_adjust_shell(self):
        conf = SysConf("".splitlines())
        conf["B"] = " $(time)"
        contents = str(conf)
        assert "B= $(time)" == contents

    def test_parse_empty(self):
        contents = ""
        conf = SysConf(contents.splitlines())
        assert "" == str(conf).strip()

    def test_parse_add_new(self):
        contents = "BLAH=b"
        conf = SysConf(contents.splitlines())
        conf["Z"] = "d"
        contents = str(conf)
        assert "Z=d" in contents
        assert "BLAH=b" in contents



"""cloudinit.signal_handler tests"""

import re
import signal
import sys

from cloudinit.signal_handler import _handle_exit


class TestSignalHandler:
    """Test signal_handler.py"""

    def test_handle_exit(self, mocker, caplog):
        """Test handle_exit()"""
        mocker.patch("cloudinit.signal_handler.sys.exit")
        mocker.patch("cloudinit.log.log_util.write_to_console")

        frame = sys._getframe()
        _handle_exit(signal.Signals.SIGHUP, frame)

        record = caplog.records[0]
        assert record.levelname == "INFO"
        assert re.match(
            (
                r"Received signal SIGHUP resulting in exit. Cause:\n"
                r"  Filename:.*test_signal_handler.py\n"
                r"  Function: test_handle_exit\n"
                r"  Line number: \d+"
            ),
            record.message,
        )



# This file is part of cloud-init. See LICENSE file for license information.
# pylint: disable=attribute-defined-outside-init

import random
import tempfile
from contextlib import ExitStack

import pytest

from cloudinit.config import cc_disk_setup
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import mock, skipUnlessJsonSchema


class TestIsDiskUsed:
    def setup_method(self):
        self.patches = ExitStack()
        mod_name = "cloudinit.config.cc_disk_setup"
        self.enumerate_disk = self.patches.enter_context(
            mock.patch("{0}.enumerate_disk".format(mod_name))
        )
        self.check_fs = self.patches.enter_context(
            mock.patch("{0}.check_fs".format(mod_name))
        )

    def teardown_method(self):
        self.patches.close()

    def test_multiple_child_nodes_returns_true(self):
        self.enumerate_disk.return_value = (mock.MagicMock() for _ in range(2))
        self.check_fs.return_value = (mock.MagicMock(), None, mock.MagicMock())
        assert cc_disk_setup.is_disk_used(mock.MagicMock())

    def test_valid_filesystem_returns_true(self):
        self.enumerate_disk.return_value = (mock.MagicMock() for _ in range(1))
        self.check_fs.return_value = (
            mock.MagicMock(),
            "ext4",
            mock.MagicMock(),
        )
        assert cc_disk_setup.is_disk_used(mock.MagicMock())

    def test_one_child_nodes_and_no_fs_returns_false(self):
        self.enumerate_disk.return_value = (mock.MagicMock() for _ in range(1))
        self.check_fs.return_value = (mock.MagicMock(), None, mock.MagicMock())
        assert not cc_disk_setup.is_disk_used(mock.MagicMock())


class TestGetMbrHddSize:
    def _test_for_sector_size(self, sector_size):
        size_in_bytes = random.randint(10000, 10000000) * 512
        size_in_sectors = size_in_bytes / sector_size

        def _subp(cmd, *args, **kwargs):
            assert 3 == len(cmd)
            if "--getsize64" in cmd:
                return size_in_bytes, None
            elif "--getss" in cmd:
                return sector_size, None
            raise RuntimeError("Unexpected blockdev command called")

        with mock.patch.object(cc_disk_setup.subp, "subp", _subp):
            assert size_in_sectors == cc_disk_setup.get_hdd_size("/dev/sda1")

    def test_size_for_512_byte_sectors(self):
        self._test_for_sector_size(512)

    def test_size_for_1024_byte_sectors(self):
        self._test_for_sector_size(1024)

    def test_size_for_2048_byte_sectors(self):
        self._test_for_sector_size(2048)

    def test_size_for_4096_byte_sectors(self):
        self._test_for_sector_size(4096)


class TestGetPartitionMbrLayout:
    def test_single_partition_using_boolean(self):
        assert ",,83" == cc_disk_setup.get_partition_mbr_layout(1000, True)

    def test_single_partition_using_list(self):
        disk_size = random.randint(1000000, 1000000000000)
        assert ",,83" == cc_disk_setup.get_partition_mbr_layout(
            disk_size, [100]
        )

    def test_half_and_half(self):
        disk_size = random.randint(1000000, 1000000000000)
        expected_partition_size = int(float(disk_size) / 2)
        assert ",{0},83\n,,83".format(
            expected_partition_size
        ) == cc_disk_setup.get_partition_mbr_layout(disk_size, [50, 50])

    def test_thirds_with_different_partition_type(self):
        disk_size = random.randint(1000000, 1000000000000)
        expected_partition_size = int(float(disk_size) * 0.33)
        assert ",{0},83\n,,82".format(
            expected_partition_size
        ) == cc_disk_setup.get_partition_mbr_layout(disk_size, [33, [66, 82]])


class TestCheckPartitionLayout:
    @mock.patch(
        "cloudinit.config.cc_disk_setup.check_partition_mbr_layout",
        return_value=["83"],
    )
    def test_simple_mbr(self, *args):
        assert cc_disk_setup.check_partition_layout("mbr", "/dev/xvdb1", True)
        assert cc_disk_setup.check_partition_layout(
            "mbr", "/dev/xvdb1", [(100, 83)]
        )

    @mock.patch(
        "cloudinit.config.cc_disk_setup.check_partition_gpt_layout",
        return_value=["8300"],
    )
    def test_simple1_gpt(self, *args):
        assert cc_disk_setup.check_partition_layout(
            "gpt", "/dev/xvdb1", [(100, 83)]
        )
        assert cc_disk_setup.check_partition_layout(
            "gpt", "/dev/xvdb1", [(100, 8300)]
        )
        assert (
            cc_disk_setup.check_partition_layout(
                "gpt", "/dev/xvdb1", [(100, 8301)]
            )
            is False
        )
        Linux_GUID = "0FC63DAF-8483-4772-8E79-3D69D8477DE4"
        assert cc_disk_setup.check_partition_layout(
            "gpt", "/dev/xvdb1", [(100, Linux_GUID)]
        )


class TestUpdateFsSetupDevices:
    def test_regression_1634678(self):
        # Cf. https://bugs.launchpad.net/cloud-init/+bug/1634678
        fs_setup = {
            "partition": "auto",
            "device": "/dev/xvdb1",
            "overwrite": False,
            "label": "test",
            "filesystem": "ext4",
        }

        cc_disk_setup.update_fs_setup_devices(
            [fs_setup], lambda device: device
        )

        assert {
            "_origname": "/dev/xvdb1",
            "partition": "auto",
            "device": "/dev/xvdb1",
            "overwrite": False,
            "label": "test",
            "filesystem": "ext4",
        } == fs_setup

    def test_dotted_devname(self):
        fs_setup = {
            "partition": "auto",
            "device": "ephemeral0.0",
            "label": "test2",
            "filesystem": "xfs",
        }

        cc_disk_setup.update_fs_setup_devices(
            [fs_setup], lambda device: device
        )

        assert {
            "_origname": "ephemeral0.0",
            "_partition": "auto",
            "partition": "0",
            "device": "ephemeral0",
            "label": "test2",
            "filesystem": "xfs",
        } == fs_setup

    def test_dotted_devname_populates_partition(self):
        fs_setup = {
            "device": "ephemeral0.1",
            "label": "test2",
            "filesystem": "xfs",
        }
        cc_disk_setup.update_fs_setup_devices(
            [fs_setup], lambda device: device
        )
        assert {
            "_origname": "ephemeral0.1",
            "device": "ephemeral0",
            "partition": "1",
            "label": "test2",
            "filesystem": "xfs",
        } == fs_setup


class TestPurgeDisk:
    @mock.patch(
        "cloudinit.config.cc_disk_setup.read_parttbl", return_value=None
    )
    def test_purge_disk_ptable(self, *args):
        pseudo_device = tempfile.NamedTemporaryFile()

        cc_disk_setup.purge_disk_ptable(pseudo_device.name)

        with pseudo_device as f:
            actual = f.read()

        expected = b"\0" * (1024 * 1024)

        assert expected == actual


@mock.patch(
    "cloudinit.config.cc_disk_setup.assert_and_settle_device",
    return_value=None,
)
@mock.patch(
    "cloudinit.config.cc_disk_setup.find_device_node",
    return_value=("/dev/xdb1", False),
)
@mock.patch("cloudinit.config.cc_disk_setup.device_type", return_value=None)
@mock.patch("cloudinit.config.cc_disk_setup.subp.subp", return_value=("", ""))
class TestMkfsCommandHandling:

    def test_with_cmd(
        self,
        subp,
        m_device_type,
        m_find_device,
        m_assert_and_settle_device,
        caplog,
    ):
        """mkfs honors cmd and logs warnings when extra_opts or overwrite are
        provided."""
        cc_disk_setup.mkfs(
            {
                "cmd": "mkfs -t %(filesystem)s -L %(label)s %(device)s",
                "filesystem": "ext4",
                "device": "/dev/xdb1",
                "label": "with_cmd",
                "extra_opts": ["should", "generate", "warning"],
                "overwrite": "should generate warning too",
            }
        )

        assert (
            "extra_opts "
            "ignored because cmd was specified: mkfs -t ext4 -L with_cmd "
            "/dev/xdb1" in caplog.text
        )
        assert (
            "overwrite "
            "ignored because cmd was specified: mkfs -t ext4 -L with_cmd "
            "/dev/xdb1" in caplog.text
        )

        subp.assert_called_once_with(
            "mkfs -t ext4 -L with_cmd /dev/xdb1", shell=True
        )

    @mock.patch("cloudinit.config.cc_disk_setup.subp.which")
    def test_overwrite_and_extra_opts_without_cmd(self, m_which, subp, *args):
        """mkfs observes extra_opts and overwrite settings when cmd is not
        present."""
        m_which.side_effect = lambda p: {"mkfs.ext4": "/sbin/mkfs.ext4"}[p]
        cc_disk_setup.mkfs(
            {
                "filesystem": "ext4",
                "device": "/dev/xdb1",
                "label": "without_cmd",
                "extra_opts": ["are", "added"],
                "overwrite": True,
            }
        )

        subp.assert_called_once_with(
            [
                "/sbin/mkfs.ext4",
                "-L",
                "without_cmd",
                "-F",
                "are",
                "added",
                "/dev/xdb1",
            ],
            shell=False,
        )

    @mock.patch("cloudinit.config.cc_disk_setup.subp.which")
    def test_mkswap(self, m_which, subp, *args):
        """mkfs observes extra_opts and overwrite settings when cmd is not
        present."""
        m_which.side_effect = iter([None, "/sbin/mkswap"])
        cc_disk_setup.mkfs(
            {
                "filesystem": "swap",
                "device": "/dev/xdb1",
                "label": "swap",
                "overwrite": True,
            }
        )

        assert [
            mock.call("mkfs.swap"),
            mock.call("mkswap"),
        ] == m_which.call_args_list
        subp.assert_called_once_with(
            ["/sbin/mkswap", "-L", "swap", "-f", "/dev/xdb1"], shell=False
        )


@skipUnlessJsonSchema()
class TestDebugSchema:
    """Directly test schema rather than through handle."""

    @pytest.mark.parametrize(
        "config, error_msg",
        (
            # Valid schemas tested by meta.examples in test_schema
            # Invalid schemas
            ({"disk_setup": 1}, "disk_setup: 1 is not of type 'object'"),
            ({"fs_setup": 1}, "fs_setup: 1 is not of type 'array'"),
            (
                {"device_aliases": 1},
                "device_aliases: 1 is not of type 'object'",
            ),
        ),
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        """Assert expected schema validation and error messages."""
        # New-style schema $defs exist in config/cloud-init-schema*.json
        schema = get_schema()
        with pytest.raises(SchemaValidationError, match=error_msg):
            validate_cloudconfig_schema(config, schema, strict=True)

    @pytest.mark.parametrize(
        "config",
        (
            (
                {
                    "disk_setup": {
                        "/dev/disk/by-id/google-home": {
                            "table_type": "gpt",
                            "layout": [
                                [100, "933AC7E1-2EB4-4F13-B844-0E14E2AEF915"]
                            ],
                        }
                    }
                }
            ),
        ),
    )
    @skipUnlessJsonSchema()
    def test_valid_schema(self, config):
        """Assert expected schema validation and no error messages."""
        schema = get_schema()
        validate_cloudconfig_schema(config, schema, strict=True)



# This file is part of cloud-init. See LICENSE file for license information.
"""Tests for `cloud-init analyze`"""
import pytest

from cloudinit.distros import uses_systemd
from tests.integration_tests.integration_settings import PLATFORM
from tests.integration_tests.util import get_datetime_from_string


class TestAnalyzeCommand:
    @pytest.mark.skipif(not uses_systemd(), reason="Relies on systemd output")
    @pytest.mark.skipif(
        PLATFORM != "lxd_container",
        reason="Testing lxdcontainer-specific behavior",
    )
    def test_analyze_boot_ordered_timestamps(self, module_client):
        """
        Confirm that analyze boot is working correctly in lxd containers
        and that the correct zero-point is used for the monotonic clock used
        to determine when cloud-init was activated by systemd
        """
        assert module_client.execute("cloud-init status --wait --long").ok
        result = module_client.execute("cloud-init analyze boot")
        assert result.stderr == "container"

        container_start_time = get_datetime_from_string(
            result.stdout, "^\\s*Container started at: (.+?)$"
        )
        cloudinit_activation_time = get_datetime_from_string(
            result.stdout, "^\\s*Cloud-init activated by systemd at: (.+?)$"
        )
        cloudinit_start_time = get_datetime_from_string(
            result.stdout, "^\\s*Cloud-init start: (.+?)$"
        )
        assert container_start_time < cloudinit_activation_time
        assert cloudinit_activation_time < cloudinit_start_time



# This file is part of cloud-init. See LICENSE file for license information.

import json
import os
from contextlib import ExitStack
from copy import copy, deepcopy

import pytest

from cloudinit import helpers, settings, util
from cloudinit.net import eni, network_state
from cloudinit.sources import DataSourceConfigDrive as ds
from cloudinit.sources.helpers import openstack
from tests.unittests.helpers import mock, populate_dir

PUBKEY = "ssh-rsa AAAAB3NzaC1....sIkJhq8wdX+4I3A4cYbYP ubuntu@server-460\n"
EC2_META = {
    "ami-id": "ami-00000001",
    "ami-launch-index": 0,
    "ami-manifest-path": "FIXME",
    "block-device-mapping": {
        "ami": "sda1",
        "ephemeral0": "sda2",
        "root": "/dev/sda1",
        "swap": "sda3",
    },
    "hostname": "sm-foo-test.novalocal",
    "instance-action": "none",
    "instance-id": "i-00000001",
    "instance-type": "m1.tiny",
    "local-hostname": "sm-foo-test.novalocal",
    "local-ipv4": None,
    "placement": {"availability-zone": "nova"},
    "public-hostname": "sm-foo-test.novalocal",
    "public-ipv4": "",
    "public-keys": {"0": {"openssh-key": PUBKEY}},
    "reservation-id": "r-iru5qm4m",
    "security-groups": ["default"],
}
USER_DATA = b"#!/bin/sh\necho This is user data\n"
VENDOR_DATA = {
    "magic": "",
}
VENDOR_DATA2 = {
    "static": "",
}
OSTACK_META = {
    "availability_zone": "nova",
    "files": [
        {"content_path": "/content/0000", "path": "/etc/foo.cfg"},
        {"content_path": "/content/0001", "path": "/etc/bar/bar.cfg"},
    ],
    "hostname": "sm-foo-test.novalocal",
    "meta": {"dsmode": "local", "my-meta": "my-value"},
    "name": "sm-foo-test",
    "public_keys": {"mykey": PUBKEY},
    "uuid": "b0fa911b-69d4-4476-bbe2-1c92bff6535c",
}

CONTENT_0 = b"This is contents of /etc/foo.cfg\n"
CONTENT_1 = b"# this is /etc/bar/bar.cfg\n"
NETWORK_DATA = {
    "services": [
        {"type": "dns", "address": "199.204.44.24"},
        {"type": "dns", "address": "199.204.47.54"},
    ],
    "links": [
        {
            "vif_id": "2ecc7709-b3f7-4448-9580-e1ec32d75bbd",
            "ethernet_mac_address": "fa:16:3e:69:b0:58",
            "type": "ovs",
            "mtu": None,
            "id": "tap2ecc7709-b3",
        },
        {
            "vif_id": "2f88d109-5b57-40e6-af32-2472df09dc33",
            "ethernet_mac_address": "fa:16:3e:d4:57:ad",
            "type": "ovs",
            "mtu": None,
            "id": "tap2f88d109-5b",
        },
        {
            "vif_id": "1a5382f8-04c5-4d75-ab98-d666c1ef52cc",
            "ethernet_mac_address": "fa:16:3e:05:30:fe",
            "type": "ovs",
            "mtu": None,
            "id": "tap1a5382f8-04",
            "name": "nic0",
        },
    ],
    "networks": [
        {
            "link": "tap2ecc7709-b3",
            "type": "ipv4_dhcp",
            "network_id": "6d6357ac-0f70-4afa-8bd7-c274cc4ea235",
            "id": "network0",
        },
        {
            "link": "tap2f88d109-5b",
            "type": "ipv4_dhcp",
            "network_id": "d227a9b3-6960-4d94-8976-ee5788b44f54",
            "id": "network1",
        },
        {
            "link": "tap1a5382f8-04",
            "type": "ipv4_dhcp",
            "network_id": "dab2ba57-cae2-4311-a5ed-010b263891f5",
            "id": "network2",
        },
    ],
}

NETWORK_DATA_2 = {
    "services": [
        {"type": "dns", "address": "1.1.1.191"},
        {"type": "dns", "address": "1.1.1.4"},
    ],
    "networks": [
        {
            "network_id": "d94bbe94-7abc-48d4-9c82-4628ea26164a",
            "type": "ipv4",
            "netmask": "255.255.255.248",
            "link": "eth0",
            "routes": [
                {
                    "netmask": "0.0.0.0",
                    "network": "0.0.0.0",
                    "gateway": "2.2.2.9",
                }
            ],
            "ip_address": "2.2.2.10",
            "id": "network0-ipv4",
        },
        {
            "network_id": "ca447c83-6409-499b-aaef-6ad1ae995348",
            "type": "ipv4",
            "netmask": "255.255.255.224",
            "link": "eth1",
            "routes": [],
            "ip_address": "3.3.3.24",
            "id": "network1-ipv4",
        },
    ],
    "links": [
        {
            "ethernet_mac_address": "fa:16:3e:dd:50:9a",
            "mtu": 1500,
            "type": "vif",
            "id": "eth0",
            "vif_id": "vif-foo1",
        },
        {
            "ethernet_mac_address": "fa:16:3e:a8:14:69",
            "mtu": 1500,
            "type": "vif",
            "id": "eth1",
            "vif_id": "vif-foo2",
        },
    ],
}

# This network data ha 'tap' or null type for a link.
NETWORK_DATA_3 = {
    "services": [
        {"type": "dns", "address": "172.16.36.11"},
        {"type": "dns", "address": "172.16.36.12"},
    ],
    "networks": [
        {
            "network_id": "7c41450c-ba44-401a-9ab1-1604bb2da51e",
            "type": "ipv4",
            "netmask": "255.255.255.128",
            "link": "tap77a0dc5b-72",
            "ip_address": "172.17.48.18",
            "id": "network0",
            "routes": [
                {
                    "netmask": "0.0.0.0",
                    "network": "0.0.0.0",
                    "gateway": "172.17.48.1",
                }
            ],
        },
        {
            "network_id": "7c41450c-ba44-401a-9ab1-1604bb2da51e",
            "type": "ipv6",
            "netmask": "ffff:ffff:ffff:ffff::",
            "link": "tap77a0dc5b-72",
            "ip_address": "fdb8:52d0:9d14:0:f816:3eff:fe9f:70d",
            "id": "network1",
            "routes": [
                {
                    "netmask": "::",
                    "network": "::",
                    "gateway": "fdb8:52d0:9d14::1",
                }
            ],
        },
        {
            "network_id": "1f53cb0e-72d3-47c7-94b9-ff4397c5fe54",
            "type": "ipv4",
            "netmask": "255.255.255.128",
            "link": "tap7d6b7bec-93",
            "ip_address": "172.16.48.13",
            "id": "network2",
            "routes": [
                {
                    "netmask": "0.0.0.0",
                    "network": "0.0.0.0",
                    "gateway": "172.16.48.1",
                },
                {
                    "netmask": "255.255.0.0",
                    "network": "172.16.0.0",
                    "gateway": "172.16.48.1",
                },
            ],
        },
    ],
    "links": [
        {
            "ethernet_mac_address": "fa:16:3e:dd:50:9a",
            "mtu": None,
            "type": "tap",
            "id": "tap77a0dc5b-72",
            "vif_id": "77a0dc5b-720e-41b7-bfa7-1b2ff62e0d48",
        },
        {
            "ethernet_mac_address": "fa:16:3e:a8:14:69",
            "mtu": None,
            "type": None,
            "id": "tap7d6b7bec-93",
            "vif_id": "7d6b7bec-93e6-4c03-869a-ddc5014892d5",
        },
    ],
}

BOND_MAC = "fa:16:3e:b3:72:36"
NETWORK_DATA_BOND = {
    "services": [
        {"type": "dns", "address": "1.1.1.191"},
        {"type": "dns", "address": "1.1.1.4"},
    ],
    "networks": [
        {
            "id": "network2-ipv4",
            "ip_address": "2.2.2.13",
            "link": "vlan2",
            "netmask": "255.255.255.248",
            "network_id": "4daf5ce8-38cf-4240-9f1a-04e86d7c6117",
            "type": "ipv4",
            "routes": [
                {
                    "netmask": "0.0.0.0",
                    "network": "0.0.0.0",
                    "gateway": "2.2.2.9",
                }
            ],
        },
        {
            "id": "network3-ipv4",
            "ip_address": "10.0.1.5",
            "link": "vlan3",
            "netmask": "255.255.255.248",
            "network_id": "a9e2f47c-3c43-4782-94d0-e1eeef1c8c9d",
            "type": "ipv4",
            "routes": [
                {
                    "netmask": "255.255.255.255",
                    "network": "192.168.1.0",
                    "gateway": "10.0.1.1",
                }
            ],
        },
    ],
    "links": [
        {
            "ethernet_mac_address": "0c:c4:7a:34:6e:3c",
            "id": "eth0",
            "mtu": 1500,
            "type": "phy",
        },
        {
            "ethernet_mac_address": "0c:c4:7a:34:6e:3d",
            "id": "eth1",
            "mtu": 1500,
            "type": "phy",
        },
        {
            "bond_links": ["eth0", "eth1"],
            "bond_miimon": 100,
            "bond_mode": "4",
            "bond_xmit_hash_policy": "layer3+4",
            "ethernet_mac_address": BOND_MAC,
            "id": "bond0",
            "type": "bond",
        },
        {
            "ethernet_mac_address": "fa:16:3e:b3:72:30",
            "id": "vlan2",
            "type": "vlan",
            "vlan_id": 602,
            "vlan_link": "bond0",
            "vlan_mac_address": "fa:16:3e:b3:72:30",
        },
        {
            "ethernet_mac_address": "fa:16:3e:66:ab:a6",
            "id": "vlan3",
            "type": "vlan",
            "vlan_id": 612,
            "vlan_link": "bond0",
            "vlan_mac_address": "fa:16:3e:66:ab:a6",
        },
    ],
}

NETWORK_DATA_VLAN = {
    "services": [{"type": "dns", "address": "1.1.1.191"}],
    "networks": [
        {
            "id": "network1-ipv4",
            "ip_address": "10.0.1.5",
            "link": "vlan1",
            "netmask": "255.255.255.248",
            "network_id": "a9e2f47c-3c43-4782-94d0-e1eeef1c8c9d",
            "type": "ipv4",
            "routes": [
                {
                    "netmask": "255.255.255.255",
                    "network": "192.168.1.0",
                    "gateway": "10.0.1.1",
                }
            ],
        }
    ],
    "links": [
        {
            "ethernet_mac_address": "fa:16:3e:69:b0:58",
            "id": "eth0",
            "mtu": 1500,
            "type": "phy",
        },
        {
            "ethernet_mac_address": "fa:16:3e:b3:72:30",
            "id": "vlan1",
            "type": "vlan",
            "vlan_id": 602,
            "vlan_link": "eth0",
            "vlan_mac_address": "fa:16:3e:b3:72:30",
        },
    ],
}

KNOWN_MACS = {
    "fa:16:3e:69:b0:58": "enp0s1",
    "fa:16:3e:d4:57:ad": "enp0s2",
    "fa:16:3e:dd:50:9a": "foo1",
    "fa:16:3e:a8:14:69": "foo2",
    "fa:16:3e:ed:9a:59": "foo3",
    "0c:c4:7a:34:6e:3d": "oeth1",
    "0c:c4:7a:34:6e:3c": "oeth0",
}

CFG_DRIVE_FILES_V2 = {
    "ec2/2009-04-04/meta-data.json": json.dumps(EC2_META),
    "ec2/2009-04-04/user-data": USER_DATA,
    "ec2/latest/meta-data.json": json.dumps(EC2_META),
    "ec2/latest/user-data": USER_DATA,
    "openstack/2012-08-10/meta_data.json": json.dumps(OSTACK_META),
    "openstack/2012-08-10/user_data": USER_DATA,
    "openstack/content/0000": CONTENT_0,
    "openstack/content/0001": CONTENT_1,
    "openstack/latest/meta_data.json": json.dumps(OSTACK_META),
    "openstack/latest/user_data": USER_DATA,
    "openstack/latest/vendor_data.json": json.dumps(VENDOR_DATA),
    "openstack/latest/vendor_data2.json": json.dumps(VENDOR_DATA2),
    "openstack/latest/network_data.json": json.dumps(NETWORK_DATA),
    "openstack/2015-10-15/meta_data.json": json.dumps(OSTACK_META),
    "openstack/2015-10-15/user_data": USER_DATA,
    "openstack/2015-10-15/network_data.json": json.dumps(NETWORK_DATA),
    "openstack/2015-10-15/vendor_data.json": json.dumps(VENDOR_DATA),
    "openstack/2015-10-15/vendor_data2.json": json.dumps(VENDOR_DATA2),
}

M_PATH = "cloudinit.sources.DataSourceConfigDrive."


class TestConfigDriveDataSource:
    @pytest.fixture(autouse=True)
    def fixtures(self, mocker, tmp_path):
        mocker.patch(M_PATH + "util.find_devs_with", return_value=[])
        populate_dir(str(tmp_path), CFG_DRIVE_FILES_V2)

    def test_ec2_metadata(self, tmp_path):
        found = ds.read_config_drive(tmp_path)
        assert "ec2-metadata" in found
        ec2_md = found["ec2-metadata"]
        assert EC2_META == ec2_md

    def test_dev_os_remap(self, tmp_path):
        cfg_ds = ds.DataSourceConfigDrive(
            settings.CFG_BUILTIN, None, helpers.Paths({})
        )
        found = ds.read_config_drive(tmp_path)
        cfg_ds.metadata = found["metadata"]
        name_tests = {
            "ami": "/dev/vda1",
            "root": "/dev/vda1",
            "ephemeral0": "/dev/vda2",
            "swap": "/dev/vda3",
        }
        for name, dev_name in name_tests.items():
            with ExitStack() as mocks:
                provided_name = dev_name[len("/dev/") :]
                provided_name = "s" + provided_name[1:]
                find_mock = mocks.enter_context(
                    mock.patch.object(
                        util, "find_devs_with", return_value=[provided_name]
                    )
                )
                # We want os.path.exists() to return False on its first call,
                # and True on its second call.  We use a handy generator as
                # the mock side effect for this.  The mocked function returns
                # what the side effect returns.

                def exists_side_effect():
                    yield False
                    yield True

                exists_mock = mocks.enter_context(
                    mock.patch.object(
                        os.path, "exists", side_effect=exists_side_effect()
                    )
                )
                assert dev_name == cfg_ds.device_name_to_device(name)

                find_mock.assert_called_once_with(mock.ANY)
                assert exists_mock.call_count == 2

    def test_dev_os_map(self, tmp_path):
        cfg_ds = ds.DataSourceConfigDrive(
            settings.CFG_BUILTIN, None, helpers.Paths({})
        )
        found = ds.read_config_drive(tmp_path)
        os_md = found["metadata"]
        cfg_ds.metadata = os_md
        name_tests = {
            "ami": "/dev/vda1",
            "root": "/dev/vda1",
            "ephemeral0": "/dev/vda2",
            "swap": "/dev/vda3",
        }
        for name, dev_name in name_tests.items():
            with ExitStack() as mocks:
                find_mock = mocks.enter_context(
                    mock.patch.object(
                        util, "find_devs_with", return_value=[dev_name]
                    )
                )
                exists_mock = mocks.enter_context(
                    mock.patch.object(os.path, "exists", return_value=True)
                )
                assert dev_name == cfg_ds.device_name_to_device(name)

                find_mock.assert_called_once_with(mock.ANY)
                exists_mock.assert_called_once_with(mock.ANY)

    def test_dev_ec2_remap(self, tmp_path):
        cfg_ds = ds.DataSourceConfigDrive(
            settings.CFG_BUILTIN, None, helpers.Paths({})
        )
        found = ds.read_config_drive(tmp_path)
        ec2_md = found["ec2-metadata"]
        os_md = found["metadata"]
        cfg_ds.ec2_metadata = ec2_md
        cfg_ds.metadata = os_md
        name_tests = {
            "ami": "/dev/vda1",
            "root": "/dev/vda1",
            "ephemeral0": "/dev/vda2",
            "swap": "/dev/vda3",
            None: None,
            "bob": None,
            "root2k": None,
        }
        for name, dev_name in name_tests.items():
            # We want os.path.exists() to return False on its first call,
            # and True on its second call.  We use a handy generator as
            # the mock side effect for this.  The mocked function returns
            # what the side effect returns.
            def exists_side_effect():
                yield False
                yield True

            with mock.patch.object(
                os.path, "exists", side_effect=exists_side_effect()
            ):
                assert dev_name == cfg_ds.device_name_to_device(name)
                # We don't assert the call count for os.path.exists() because
                # not all of the entries in name_tests results in two calls to
                # that function.  Specifically, 'root2k' doesn't seem to call
                # it at all.

    def test_dev_ec2_map(self, tmp_path):
        cfg_ds = ds.DataSourceConfigDrive(
            settings.CFG_BUILTIN, None, helpers.Paths({})
        )
        found = ds.read_config_drive(tmp_path)
        ec2_md = found["ec2-metadata"]
        os_md = found["metadata"]
        cfg_ds.ec2_metadata = ec2_md
        cfg_ds.metadata = os_md
        name_tests = {
            "ami": "/dev/sda1",
            "root": "/dev/sda1",
            "ephemeral0": "/dev/sda2",
            "swap": "/dev/sda3",
            None: None,
            "bob": None,
            "root2k": None,
        }
        for name, dev_name in name_tests.items():
            with mock.patch.object(os.path, "exists", return_value=True):
                assert dev_name == cfg_ds.device_name_to_device(name)

    def test_dir_valid(self, tmp_path):
        """Verify a dir is read as such."""

        found = ds.read_config_drive(tmp_path)

        expected_md = copy(OSTACK_META)
        expected_md["instance-id"] = expected_md["uuid"]
        expected_md["local-hostname"] = expected_md["hostname"]

        assert USER_DATA == found["userdata"]
        assert expected_md == found["metadata"]
        assert NETWORK_DATA == found["networkdata"]
        assert VENDOR_DATA == found["vendordata"]
        assert VENDOR_DATA2 == found["vendordata2"]
        assert found["files"]["/etc/foo.cfg"] == CONTENT_0
        assert found["files"]["/etc/bar/bar.cfg"] == CONTENT_1

    def test_seed_dir_valid_extra(self, tmp_path):
        """Verify extra files do not affect datasource validity."""

        data = copy(CFG_DRIVE_FILES_V2)
        data["myfoofile.txt"] = "myfoocontent"
        data["openstack/latest/random-file.txt"] = "random-content"

        populate_dir(str(tmp_path), data)

        found = ds.read_config_drive(tmp_path)

        expected_md = copy(OSTACK_META)
        expected_md["instance-id"] = expected_md["uuid"]
        expected_md["local-hostname"] = expected_md["hostname"]

        assert expected_md == found["metadata"]

    def test_seed_dir_bad_json_metadata(self, tmp_path):
        """Verify that bad json in metadata raises BrokenConfigDriveDir."""
        data = copy(CFG_DRIVE_FILES_V2)

        data["openstack/2012-08-10/meta_data.json"] = "non-json garbage {}"
        data["openstack/2015-10-15/meta_data.json"] = "non-json garbage {}"
        data["openstack/latest/meta_data.json"] = "non-json garbage {}"

        populate_dir(str(tmp_path), data)

        with pytest.raises(openstack.BrokenMetadata):
            ds.read_config_drive(tmp_path)

    def test_seed_dir_no_configdrive(self, tmp_path):
        """Verify that no metadata raises NonConfigDriveDir."""

        my_d = os.path.join(tmp_path, "non-configdrive")
        data = copy(CFG_DRIVE_FILES_V2)
        data["myfoofile.txt"] = "myfoocontent"
        data["openstack/latest/random-file.txt"] = "random-content"
        data["content/foo"] = "foocontent"

        with pytest.raises(openstack.NonReadable):
            ds.read_config_drive(my_d)

    def test_seed_dir_missing(self, tmp_path):
        """Verify that missing seed_dir raises NonConfigDriveDir."""
        my_d = os.path.join(tmp_path, "nonexistantdirectory")
        with pytest.raises(openstack.NonReadable):
            ds.read_config_drive(my_d)

    def test_find_candidates(self):
        devs_with_answers = {}

        def my_devs_with(*args, **kwargs):
            criteria = args[0] if args else kwargs.pop("criteria", None)
            return devs_with_answers.get(criteria, [])

        def my_is_partition(dev):
            return dev[-1] in "0123456789" and not dev.startswith("sr")

        orig_find_devs_with = util.find_devs_with
        orig_is_partition = util.is_partition
        try:
            util.find_devs_with = my_devs_with
            util.is_partition = my_is_partition

            devs_with_answers = {
                "TYPE=vfat": [],
                "TYPE=iso9660": ["/dev/vdb"],
                "LABEL=config-2": ["/dev/vdb"],
            }
            assert ["/dev/vdb"] == ds.find_candidate_devs()

            # add a vfat item
            # zdd reverse sorts after vdb, but config-2 label is preferred
            devs_with_answers["TYPE=vfat"] = ["/dev/zdd"]
            assert ["/dev/vdb", "/dev/zdd"] == ds.find_candidate_devs()

            # verify that partitions are considered, that have correct label.
            devs_with_answers = {
                "TYPE=vfat": ["/dev/sda1"],
                "TYPE=iso9660": [],
                "LABEL=config-2": ["/dev/vdb3"],
            }
            assert ["/dev/vdb3"] == ds.find_candidate_devs()

            # Verify that uppercase labels are also found.
            devs_with_answers = {
                "TYPE=vfat": [],
                "TYPE=iso9660": ["/dev/vdb"],
                "LABEL=CONFIG-2": ["/dev/vdb"],
            }
            assert ["/dev/vdb"] == ds.find_candidate_devs()

        finally:
            util.find_devs_with = orig_find_devs_with
            util.is_partition = orig_is_partition

    @mock.patch(M_PATH + "on_first_boot")
    def test_pubkeys_v2(self, on_first_boot, tmp_path):
        """Verify that public-keys work in config-drive-v2."""
        myds = cfg_ds_from_dir(tmp_path, files=CFG_DRIVE_FILES_V2)
        assert myds.get_public_ssh_keys() == [
            OSTACK_META["public_keys"]["mykey"]
        ]
        assert "configdrive" == myds.cloud_name
        assert "openstack" == myds.platform
        assert "seed-dir (%s/seed)" % tmp_path == myds.subplatform

    def test_subplatform_config_drive_when_starts_with_dev(self):
        """subplatform reports config-drive when source starts with /dev/."""
        cfg_ds = ds.DataSourceConfigDrive(
            settings.CFG_BUILTIN, None, helpers.Paths({})
        )
        with mock.patch(M_PATH + "find_candidate_devs") as m_find_devs:
            with mock.patch(M_PATH + "util.mount_cb"):
                with mock.patch(M_PATH + "on_first_boot"):
                    m_find_devs.return_value = ["/dev/anything"]
                    assert True is cfg_ds.get_data()
        assert "config-disk (/dev/anything)" == cfg_ds.subplatform


@mock.patch(
    "cloudinit.net.is_openvswitch_internal_interface",
    mock.Mock(return_value=False),
)
class TestNetJson:
    @mock.patch(M_PATH + "on_first_boot")
    def test_network_data_is_found(self, on_first_boot, tmp_path):
        """Verify that network_data is present in ds in config-drive-v2."""
        myds = cfg_ds_from_dir(tmp_path, files=CFG_DRIVE_FILES_V2)
        assert myds.network_json is not None

    @mock.patch(M_PATH + "on_first_boot")
    def test_network_config_is_converted(self, on_first_boot, tmp_path):
        """Verify that network_data is converted and present on ds object."""
        myds = cfg_ds_from_dir(tmp_path, files=CFG_DRIVE_FILES_V2)
        network_config = openstack.convert_net_json(
            NETWORK_DATA, known_macs=KNOWN_MACS
        )
        assert myds.network_config == network_config

    def test_network_config_conversion_dhcp6(self):
        """Test some ipv6 input network json and check the expected
        conversions."""
        in_data = {
            "links": [
                {
                    "vif_id": "2ecc7709-b3f7-4448-9580-e1ec32d75bbd",
                    "ethernet_mac_address": "fa:16:3e:69:b0:58",
                    "type": "ovs",
                    "mtu": None,
                    "id": "tap2ecc7709-b3",
                },
                {
                    "vif_id": "2f88d109-5b57-40e6-af32-2472df09dc33",
                    "ethernet_mac_address": "fa:16:3e:d4:57:ad",
                    "type": "ovs",
                    "mtu": None,
                    "id": "tap2f88d109-5b",
                },
            ],
            "networks": [
                {
                    "link": "tap2ecc7709-b3",
                    "type": "ipv6_dhcpv6-stateless",
                    "network_id": "6d6357ac-0f70-4afa-8bd7-c274cc4ea235",
                    "id": "network0",
                },
                {
                    "link": "tap2f88d109-5b",
                    "type": "ipv6_dhcpv6-stateful",
                    "network_id": "d227a9b3-6960-4d94-8976-ee5788b44f54",
                    "id": "network1",
                },
            ],
        }
        out_data = {
            "version": 1,
            "config": [
                {
                    "mac_address": "fa:16:3e:69:b0:58",
                    "mtu": None,
                    "name": "enp0s1",
                    "subnets": [{"type": "ipv6_dhcpv6-stateless"}],
                    "type": "physical",
                },
                {
                    "mac_address": "fa:16:3e:d4:57:ad",
                    "mtu": None,
                    "name": "enp0s2",
                    "subnets": [{"type": "ipv6_dhcpv6-stateful"}],
                    "type": "physical",
                    "accept-ra": True,
                },
            ],
        }
        conv_data = openstack.convert_net_json(in_data, known_macs=KNOWN_MACS)
        assert out_data == conv_data

    def test_network_config_conversions(self):
        """Tests a bunch of input network json and checks the
        expected conversions."""
        in_datas = [
            NETWORK_DATA,
            {
                "services": [{"type": "dns", "address": "172.19.0.12"}],
                "networks": [
                    {
                        "network_id": "dacd568d-5be6-4786-91fe-750c374b78b4",
                        "type": "ipv4",
                        "netmask": "255.255.252.0",
                        "link": "tap1a81968a-79",
                        "routes": [
                            {
                                "netmask": "0.0.0.0",
                                "network": "0.0.0.0",
                                "gateway": "172.19.3.254",
                            }
                        ],
                        "ip_address": "172.19.1.34",
                        "id": "network0",
                    }
                ],
                "links": [
                    {
                        "type": "bridge",
                        "vif_id": "1a81968a-797a-400f-8a80-567f997eb93f",
                        "ethernet_mac_address": "fa:16:3e:ed:9a:59",
                        "id": "tap1a81968a-79",
                        "mtu": None,
                    }
                ],
            },
        ]
        out_datas = [
            {
                "version": 1,
                "config": [
                    {
                        "subnets": [{"type": "dhcp4"}],
                        "type": "physical",
                        "mac_address": "fa:16:3e:69:b0:58",
                        "name": "enp0s1",
                        "mtu": None,
                    },
                    {
                        "subnets": [{"type": "dhcp4"}],
                        "type": "physical",
                        "mac_address": "fa:16:3e:d4:57:ad",
                        "name": "enp0s2",
                        "mtu": None,
                    },
                    {
                        "subnets": [{"type": "dhcp4"}],
                        "type": "physical",
                        "mac_address": "fa:16:3e:05:30:fe",
                        "name": "nic0",
                        "mtu": None,
                    },
                    {
                        "type": "nameserver",
                        "address": "199.204.44.24",
                    },
                    {
                        "type": "nameserver",
                        "address": "199.204.47.54",
                    },
                ],
            },
            {
                "version": 1,
                "config": [
                    {
                        "name": "foo3",
                        "mac_address": "fa:16:3e:ed:9a:59",
                        "mtu": None,
                        "type": "physical",
                        "subnets": [
                            {
                                "address": "172.19.1.34",
                                "netmask": "255.255.252.0",
                                "type": "static",
                                "ipv4": True,
                                "routes": [
                                    {
                                        "gateway": "172.19.3.254",
                                        "netmask": "0.0.0.0",
                                        "network": "0.0.0.0",
                                    }
                                ],
                            }
                        ],
                    },
                    {
                        "type": "nameserver",
                        "address": "172.19.0.12",
                    },
                ],
            },
        ]
        for in_data, out_data in zip(in_datas, out_datas):
            conv_data = openstack.convert_net_json(
                in_data, known_macs=KNOWN_MACS
            )
            assert out_data == conv_data


@mock.patch(
    "cloudinit.net.is_openvswitch_internal_interface",
    mock.Mock(return_value=False),
)
class TestConvertNetworkData:
    def _getnames_in_config(self, ncfg):
        return set(
            [n["name"] for n in ncfg["config"] if n["type"] == "physical"]
        )

    def test_conversion_fills_names(self):
        ncfg = openstack.convert_net_json(NETWORK_DATA, known_macs=KNOWN_MACS)
        expected = set(["nic0", "enp0s1", "enp0s2"])
        found = self._getnames_in_config(ncfg)
        assert found == expected

    @mock.patch("cloudinit.net.get_interfaces_by_mac")
    def test_convert_reads_system_prefers_name(self, get_interfaces_by_mac):
        macs = KNOWN_MACS.copy()
        macs.update(
            {"fa:16:3e:05:30:fe": "foonic1", "fa:16:3e:69:b0:58": "ens1"}
        )
        get_interfaces_by_mac.return_value = macs

        ncfg = openstack.convert_net_json(NETWORK_DATA)
        expected = set(["nic0", "ens1", "enp0s2"])
        found = self._getnames_in_config(ncfg)
        assert found == expected

    def test_convert_raises_value_error_on_missing_name(self):
        macs = {"aa:aa:aa:aa:aa:00": "ens1"}
        with mock.patch(
            "cloudinit.sources.helpers.openstack.util.udevadm_settle"
        ):
            with pytest.raises(ValueError):
                openstack.convert_net_json(
                    NETWORK_DATA,
                    known_macs=macs,
                )

    @mock.patch("cloudinit.subp.which")
    def test_conversion_with_route(self, m_which, tmp_path):
        ncfg = openstack.convert_net_json(
            NETWORK_DATA_2, known_macs=KNOWN_MACS
        )
        # not the best test, but see that we get a route in the
        # network config and that it gets rendered to an ENI file
        routes = []
        for n in ncfg["config"]:
            for s in n.get("subnets", []):
                routes.extend(s.get("routes", []))
        assert {
            "network": "0.0.0.0",
            "netmask": "0.0.0.0",
            "gateway": "2.2.2.9",
        } in routes

        m_which.return_value = "/sbin/ip"
        eni_renderer = eni.Renderer()
        eni_renderer.render_network_state(
            network_state.parse_net_config_data(ncfg), target=str(tmp_path)
        )
        with open(
            os.path.join(tmp_path, "etc", "network", "interfaces"), "r"
        ) as f:
            eni_rendering = f.read()
            assert "ip route add default via 2.2.2.9" in eni_rendering

        m_which.return_value = None
        eni_renderer = eni.Renderer()
        eni_renderer.render_network_state(
            network_state.parse_net_config_data(ncfg), target=str(tmp_path)
        )
        with open(
            os.path.join(tmp_path, "etc", "network", "interfaces"), "r"
        ) as f:
            eni_rendering = f.read()
            assert "route add default gw 2.2.2.9" in eni_rendering

    def test_conversion_with_tap(self):
        ncfg = openstack.convert_net_json(
            NETWORK_DATA_3, known_macs=KNOWN_MACS
        )
        physicals = set()
        for i in ncfg["config"]:
            if i.get("type") == "physical":
                physicals.add(i["name"])
        assert physicals == set(("foo1", "foo2"))

    def test_bond_conversion(self, tmp_path):
        # light testing of bond conversion and eni rendering of bond
        ncfg = openstack.convert_net_json(
            NETWORK_DATA_BOND, known_macs=KNOWN_MACS
        )
        eni_renderer = eni.Renderer()

        eni_renderer.render_network_state(
            network_state.parse_net_config_data(ncfg), target=str(tmp_path)
        )
        with open(
            os.path.join(tmp_path, "etc", "network", "interfaces"), "r"
        ) as f:
            eni_rendering = f.read()

        # Verify there are expected interfaces in the net config.
        interfaces = sorted(
            [
                i["name"]
                for i in ncfg["config"]
                if i["type"] in ("vlan", "bond", "physical")
            ]
        )
        assert (
            sorted(["oeth0", "oeth1", "bond0", "bond0.602", "bond0.612"])
            == interfaces
        )

        words = eni_rendering.split()
        # 'eth0' and 'eth1' are the ids. because their mac addresses
        # map to other names, we should not see them in the ENI
        assert "eth0" not in words
        assert "eth1" not in words

        # oeth0 and oeth1 are the interface names for eni.
        # bond0 will be generated for the bond. Each should be auto.
        assert "auto oeth0" in eni_rendering
        assert "auto oeth1" in eni_rendering
        assert "auto bond0" in eni_rendering
        # The bond should have the given mac address
        pos = eni_rendering.find("auto bond0")
        assert BOND_MAC in eni_rendering[pos:]

    def test_vlan(self, tmp_path):
        # light testing of vlan config conversion and eni rendering
        ncfg = openstack.convert_net_json(
            NETWORK_DATA_VLAN, known_macs=KNOWN_MACS
        )
        eni_renderer = eni.Renderer()
        eni_renderer.render_network_state(
            network_state.parse_net_config_data(ncfg), target=str(tmp_path)
        )
        with open(
            os.path.join(tmp_path, "etc", "network", "interfaces"), "r"
        ) as f:
            eni_rendering = f.read()

        assert "iface enp0s1" in eni_rendering
        assert "address 10.0.1.5" in eni_rendering
        assert "auto enp0s1.602" in eni_rendering

    def test_mac_addrs_can_be_upper_case(self):
        # input mac addresses on rackspace may be upper case
        my_netdata = deepcopy(NETWORK_DATA)
        for link in my_netdata["links"]:
            link["ethernet_mac_address"] = link["ethernet_mac_address"].upper()

        ncfg = openstack.convert_net_json(my_netdata, known_macs=KNOWN_MACS)
        config_name2mac = {}
        for n in ncfg["config"]:
            if n["type"] == "physical":
                config_name2mac[n["name"]] = n["mac_address"]

        expected = {
            "nic0": "fa:16:3e:05:30:fe",
            "enp0s1": "fa:16:3e:69:b0:58",
            "enp0s2": "fa:16:3e:d4:57:ad",
        }
        assert expected == config_name2mac

    def test_unknown_device_types_accepted(self, caplog):
        # If we don't recognise a link, we should treat it as physical for a
        # best-effort boot
        my_netdata = deepcopy(NETWORK_DATA)
        my_netdata["links"][0]["type"] = "my-special-link-type"

        ncfg = openstack.convert_net_json(my_netdata, known_macs=KNOWN_MACS)
        config_name2mac = {}
        for n in ncfg["config"]:
            if n["type"] == "physical":
                config_name2mac[n["name"]] = n["mac_address"]

        expected = {
            "nic0": "fa:16:3e:05:30:fe",
            "enp0s1": "fa:16:3e:69:b0:58",
            "enp0s2": "fa:16:3e:d4:57:ad",
        }
        assert expected == config_name2mac

        # We should, however, warn the user that we don't recognise the type
        assert (
            "Unknown network_data link type (my-special-link-type)"
            in caplog.text
        )


def cfg_ds_from_dir(base_d, files=None):
    run = os.path.join(base_d, "run")
    os.mkdir(run)
    cfg_ds = ds.DataSourceConfigDrive(
        settings.CFG_BUILTIN, None, helpers.Paths({"run_dir": run})
    )
    cfg_ds.seed_dir = os.path.join(base_d, "seed")
    if files:
        populate_dir(cfg_ds.seed_dir, files)
    cfg_ds.known_macs = KNOWN_MACS.copy()
    if not cfg_ds.get_data():
        raise RuntimeError(
            "Data source did not extract itself from seed directory %s"
            % cfg_ds.seed_dir
        )
    return cfg_ds


def populate_ds_from_read_config(cfg_ds, source, results):
    """Patch the DataSourceConfigDrive from the results of
    read_config_drive_dir hopefully in line with what it would have
    if cfg_ds.get_data had been successfully called"""
    cfg_ds.source = source
    cfg_ds.metadata = results.get("metadata")
    cfg_ds.ec2_metadata = results.get("ec2-metadata")
    cfg_ds.userdata_raw = results.get("userdata")
    cfg_ds.vendordata_raw = results.get("vendordata")
    cfg_ds.vendordata2_raw = results.get("vendordata2")
    cfg_ds.version = results.get("version")
    cfg_ds.network_json = results.get("networkdata")
    cfg_ds._network_config = openstack.convert_net_json(
        cfg_ds.network_json, known_macs=KNOWN_MACS
    )



# This file is part of cloud-init. See LICENSE file for license information.

import logging
import os
from types import SimpleNamespace

import pytest

from cloudinit import handlers, helpers, settings, url_helper, util
from cloudinit.cmd import main
from tests.unittests.helpers import mock


@pytest.fixture
def walker_data(tmp_path):
    data = {
        "handlercount": 0,
        "frequency": "",
        "handlerdir": str(tmp_path),
        "handlers": helpers.ContentHandlers(),
        "data": None,
    }

    expected_module_name = "part-handler-%03d" % (data["handlercount"],)
    expected_file_name = "%s.py" % expected_module_name
    expected_file_fullname = os.path.join(
        data["handlerdir"], expected_file_name
    )

    return SimpleNamespace(
        data=data,
        expected_module_name=expected_module_name,
        expected_file_fullname=expected_file_fullname,
        module_fake=FakeModule(),
        ctype=None,
        filename=None,
        payload="dummy payload",
    )


@pytest.fixture
def m_write_file():
    """Mock the write_file() function."""
    with mock.patch("cloudinit.util.write_file") as mock_write:
        yield mock_write


@pytest.fixture
def handler_data_factory():
    def _make_handler_data(frequency=settings.PER_INSTANCE):
        return SimpleNamespace(
            data="fake data",
            ctype="fake ctype",
            filename="fake filename",
            payload="fake payload",
            frequency=settings.PER_INSTANCE,
            headers={"Content-Type": "fake ctype"},
        )

    return _make_handler_data


class FakeModule(handlers.Handler):
    def __init__(self):
        handlers.Handler.__init__(self, settings.PER_ALWAYS)
        self.types = []

    def list_types(self):
        return self.types

    def handle_part(self, data, ctype, filename, payload, frequency):
        pass


class TestWalkerHandleHandler:
    def test_no_errors(self, walker_data, m_write_file):
        """Payload gets written to file and added to C{pdata}."""
        with mock.patch(
            "cloudinit.importer.import_module",
            return_value=walker_data.module_fake,
        ) as mockobj:
            handlers.walker_handle_handler(
                walker_data.data,
                walker_data.ctype,
                walker_data.filename,
                walker_data.payload,
            )
            mockobj.assert_called_once_with(walker_data.expected_module_name)
        m_write_file.assert_called_once_with(
            walker_data.expected_file_fullname, walker_data.payload, 0o600
        )
        assert walker_data.data["handlercount"] == 1

    def test_import_error(self, walker_data, m_write_file):
        """Module import errors are logged. No handler added to C{pdata}."""
        with mock.patch(
            "cloudinit.importer.import_module", side_effect=ImportError
        ) as mockobj:
            handlers.walker_handle_handler(
                walker_data.data,
                walker_data.ctype,
                walker_data.filename,
                walker_data.payload,
            )
            mockobj.assert_called_once_with(walker_data.expected_module_name)
        m_write_file.assert_called_once_with(
            walker_data.expected_file_fullname, walker_data.payload, 0o600
        )
        assert walker_data.data["handlercount"] == 0

    def test_attribute_error(self, walker_data, m_write_file):
        """Attribute errors are logged. No handler added to C{pdata}."""
        with mock.patch(
            "cloudinit.importer.import_module",
            side_effect=AttributeError,
            return_value=walker_data.module_fake,
        ) as mockobj:
            handlers.walker_handle_handler(
                walker_data.data,
                walker_data.ctype,
                walker_data.filename,
                walker_data.payload,
            )
            mockobj.assert_called_once_with(walker_data.expected_module_name)
        m_write_file.assert_called_once_with(
            walker_data.expected_file_fullname, walker_data.payload, 0o600
        )
        assert walker_data.data["handlercount"] == 0


class TestHandlerHandlePart:
    def test_normal_version_1(self, handler_data_factory):
        """
        C{handle_part} is called without C{frequency} for
        C{handler_version} == 1.
        """
        handler_data = handler_data_factory()
        mod_mock = mock.Mock(
            frequency=settings.PER_INSTANCE, handler_version=1
        )
        handlers.run_part(
            mod_mock,
            handler_data.data,
            handler_data.filename,
            handler_data.payload,
            handler_data.frequency,
            handler_data.headers,
        )
        # Assert that the handle_part() method of the mock object got
        # called with the expected arguments.
        mod_mock.handle_part.assert_called_once_with(
            handler_data.data,
            handler_data.ctype,
            handler_data.filename,
            handler_data.payload,
        )

    def test_normal_version_2(self, handler_data_factory):
        """
        C{handle_part} is called with C{frequency} for
        C{handler_version} == 2.
        """
        handler_data = handler_data_factory()
        mod_mock = mock.Mock(
            frequency=settings.PER_INSTANCE, handler_version=2
        )
        handlers.run_part(
            mod_mock,
            handler_data.data,
            handler_data.filename,
            handler_data.payload,
            handler_data.frequency,
            handler_data.headers,
        )
        # Assert that the handle_part() method of the mock object got
        # called with the expected arguments.
        mod_mock.handle_part.assert_called_once_with(
            handler_data.data,
            handler_data.ctype,
            handler_data.filename,
            handler_data.payload,
            settings.PER_INSTANCE,
        )

    def test_modfreq_per_always(self, handler_data_factory):
        """
        C{handle_part} is called regardless of frequency if nofreq is always.
        """
        handler_data = handler_data_factory(frequency=settings.PER_ONCE)
        mod_mock = mock.Mock(frequency=settings.PER_ALWAYS, handler_version=1)
        handlers.run_part(
            mod_mock,
            handler_data.data,
            handler_data.filename,
            handler_data.payload,
            handler_data.frequency,
            handler_data.headers,
        )
        # Assert that the handle_part() method of the mock object got
        # called with the expected arguments.
        mod_mock.handle_part.assert_called_once_with(
            handler_data.data,
            handler_data.ctype,
            handler_data.filename,
            handler_data.payload,
        )

    def test_no_handle_when_modfreq_once(self, handler_data_factory):
        """C{handle_part} is not called if frequency is once."""
        handler_data = handler_data_factory(frequency=settings.PER_ONCE)
        mod_mock = mock.Mock(frequency=settings.PER_ONCE)
        handlers.run_part(
            mod_mock,
            handler_data.data,
            handler_data.filename,
            handler_data.payload,
            handler_data.frequency,
            handler_data.headers,
        )
        assert 0 == mod_mock.handle_part.call_count

    def test_exception_is_caught(self, handler_data_factory):
        """Exceptions within C{handle_part} are caught and logged."""
        handler_data = handler_data_factory()
        mod_mock = mock.Mock(
            frequency=settings.PER_INSTANCE, handler_version=1
        )
        mod_mock.handle_part.side_effect = Exception
        handlers.run_part(
            mod_mock,
            handler_data.data,
            handler_data.filename,
            handler_data.payload,
            handler_data.frequency,
            handler_data.headers,
        )
        mod_mock.handle_part.assert_called_once_with(
            handler_data.data,
            handler_data.ctype,
            handler_data.filename,
            handler_data.payload,
        )


class FakeResponse:
    def __init__(self, content, status_code=200):
        self._content = content
        self._remaining_content = content
        self.status_code = status_code
        self.encoding = None

    @property
    def content(self):
        return self._remaining_content

    def iter_content(self, chunk_size, *_, **__):
        iterators = [iter(self._content)] * chunk_size
        for chunk in zip(*iterators):
            self._remaining_content = self._remaining_content[chunk_size:]
            yield bytes(chunk)


class TestCmdlineUrl:
    def test_parse_cmdline_url_nokey_raises_keyerror(self):
        with pytest.raises(KeyError):
            main.parse_cmdline_url("root=foo bar single")

    def test_parse_cmdline_url_found(self):
        cmdline = "root=foo bar single url=http://example.com arg1 -v"
        assert ("url", "http://example.com") == main.parse_cmdline_url(cmdline)

    @mock.patch("cloudinit.cmd.main.url_helper.read_file_or_url")
    def test_invalid_content(self, m_read, tmpdir):
        key = "cloud-config-url"
        url = "http://example.com/foo"
        cmdline = "ro %s=%s bar=1" % (key, url)
        m_read.return_value = url_helper.StringResponse(
            b"unexpected blob", "http://example.com/"
        )

        fpath = tmpdir.join("ccfile")
        lvl, msg = main.attempt_cmdline_url(
            fpath, network=True, cmdline=cmdline
        )
        assert logging.WARN == lvl
        assert url in msg
        assert False is os.path.exists(fpath)

    @mock.patch("cloudinit.cmd.main.url_helper.read_file_or_url")
    def test_invalid_content_url(self, m_read, tmpdir):
        key = "cloud-config-url"
        url = "http://example.com/foo"
        cmdline = "ro %s=%s bar=1" % (key, url)
        response = mock.Mock()
        response.iter_content.return_value = iter(
            (b"unexpected blob", StopIteration)
        )
        response.status_code = 200
        m_read.return_value = url_helper.UrlResponse(response)

        fpath = tmpdir.join("ccfile")
        lvl, msg = main.attempt_cmdline_url(
            fpath, network=True, cmdline=cmdline
        )
        assert logging.WARN == lvl
        assert url in msg
        assert False is os.path.exists(fpath)

    @mock.patch("cloudinit.cmd.main.url_helper.read_file_or_url")
    def test_valid_content(self, m_read, tmpdir):
        url = "http://example.com/foo"
        payload = b"#cloud-config\nmydata: foo\nbar: wark\n"
        cmdline = "ro %s=%s bar=1" % ("cloud-config-url", url)

        m_read.return_value = url_helper.StringResponse(
            payload, "http://example.com"
        )
        fpath = tmpdir.join("ccfile")
        lvl, msg = main.attempt_cmdline_url(
            fpath, network=True, cmdline=cmdline
        )
        assert util.load_binary_file(fpath) == payload
        assert logging.INFO == lvl
        assert url in msg

    @mock.patch("cloudinit.cmd.main.url_helper.read_file_or_url")
    def test_valid_content_url(self, m_read, tmpdir):
        url = "http://example.com/foo"
        payload = b"#cloud-config\nmydata: foo\nbar: wark\n"
        cmdline = "ro %s=%s bar=1" % ("cloud-config-url", url)

        response = FakeResponse(payload)
        m_read.return_value = url_helper.UrlResponse(response)

        fpath = tmpdir.join("ccfile")
        lvl, msg = main.attempt_cmdline_url(
            fpath, network=True, cmdline=cmdline
        )
        assert util.load_binary_file(fpath) == payload
        assert logging.INFO == lvl
        assert url in msg

    @mock.patch("cloudinit.cmd.main.url_helper.read_file_or_url")
    def test_no_key_found(self, m_read, tmpdir):
        cmdline = "ro mykey=http://example.com/foo root=foo"
        fpath = tmpdir.join("ccfile")
        lvl, _msg = main.attempt_cmdline_url(
            fpath, network=True, cmdline=cmdline
        )

        m_read.assert_not_called()
        assert False is os.path.exists(fpath)
        assert logging.DEBUG == lvl

    @mock.patch("cloudinit.cmd.main.url_helper.read_file_or_url")
    def test_exception_warns(self, m_read, tmpdir):
        url = "http://example.com/foo"
        cmdline = "ro cloud-config-url=%s root=LABEL=bar" % url
        fpath = tmpdir.join("ccfile")
        m_read.side_effect = url_helper.UrlError(
            cause="Unexpected Error", url="http://example.com/foo"
        )

        lvl, msg = main.attempt_cmdline_url(
            fpath, network=True, cmdline=cmdline
        )
        assert logging.WARN == lvl
        assert url in msg
        assert False is os.path.exists(fpath)



# Copyright (C) 2024-2025 Raspberry Pi Ltd. All rights reserved.
#
# Author: Paul Oberosler <paul.oberosler@raspberrypi.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import logging

from cloudinit import net, subp
from cloudinit.distros import debian

LOG = logging.getLogger(__name__)


class Distro(debian.Distro):
    def set_keymap(self, layout: str, model: str, variant: str, options: str):
        super().set_keymap(layout, model, variant, options)

        subp.subp(
            [
                "/usr/bin/raspi-config",
                "nonint",
                "update_labwc_keyboard",
            ],
        )
        subp.subp(
            [
                "/usr/bin/raspi-config",
                "nonint",
                "update_squeekboard",
                "restart",
            ],
        )
        self.manage_service("restart", "keyboard-setup")

        if subp.which("udevadm"):
            subp.subp(
                [
                    "udevadm",
                    "trigger",
                    "--subsystem-match=input",
                    "--action=change",
                ],
            )

    def apply_locale(self, locale, out_fn=None, keyname="LANG"):
        try:
            subp.subp(
                [
                    "/usr/bin/raspi-config",
                    "nonint",
                    "do_change_locale",
                    f"{locale}",
                ]
            )
        except subp.ProcessExecutionError:
            if not locale.endswith(".UTF-8"):
                LOG.info("Trying to set locale %s.UTF-8", locale)
                subp.subp(
                    [
                        "/usr/bin/raspi-config",
                        "nonint",
                        "do_change_locale",
                        f"{locale}.UTF-8",
                    ]
                )
            else:
                LOG.error("Failed to set locale %s", locale)

    def add_user(self, name, **kwargs) -> bool:
        """
        Add a user to the system using standard GNU tools

        This should be overridden on distros where useradd is not desirable or
        not available.

        Returns False if user already exists, otherwise True.
        """
        result = super().add_user(name, **kwargs)

        if not result:
            return result

        try:
            subp.subp(
                [
                    "/usr/bin/rename-user",
                    "-f",
                    "-s",
                ],
                update_env={"SUDO_USER": name},
            )

        except subp.ProcessExecutionError as e:
            LOG.error("Failed to setup user: %s", e)
            return False

        return True

    def generate_fallback_config(self):
        # Based on Photon OS implementation
        key = "disable_fallback_netcfg"
        disable_fallback_netcfg = self._cfg.get(key, True)
        LOG.debug("%s value is: %s", key, disable_fallback_netcfg)

        if not disable_fallback_netcfg:
            return net.generate_fallback_config()

        LOG.info(
            "Skipping generation of fallback network config as per "
            "configuration. Rely on Raspberry Pi OS's default "
            "network configuration."
        )
        return None



import pytest

from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.integration_settings import PLATFORM

IPV4_METADATA_ADDRESS = "169.254.169.254"
IPV6_METADATA_ADDRESS = "fd00:c1::a9fe:a9fe"


def _find_imds_read_log(log: str, address: str) -> bool:
    """
    Check if the log contains a successful IMDS read from the given address.

    Args:
        log: Contents of the cloud-init log file to search.
        address: The IMDS IP address to look for in the log.
    """
    return (
        "Successfully fetched instance metadata from IMDS at: "
        f"http://[{address}]/" in log
    )


def _test_reading_ipv6_metadata_succeeds(client: IntegrationInstance):
    """
    Test that reading metadata succeeds for an IPv6 instance.

    This function performs the following checks:
    1. Reads the cloud-init log file and ensures that a log entry indicates
         the metadata was successfully read from the IPv6 IMDS address.
    2. Verifies that the cloud-init status indicates a successful read of
    """
    log = client.read_from_file("/var/log/cloud-init.log")
    matching_line = None
    for line in log.splitlines():
        if _find_imds_read_log(line, IPV6_METADATA_ADDRESS):
            matching_line = line
            break
    assert matching_line is not None
    assert client.execute("cloud-init status --long").ok


def _test_reading_ipv6_metadata_fails(client: IntegrationInstance):
    """
    Test that reading metadata fails for an IPv6 instance and that the fallback
    datasource is used instead.
    This function performs the following checks:
    1. Reads the cloud-init log file and ensures that no log entry indicates
       an attempt to read metadata from the IPv6 IMDS address.
    2. Verifies that the cloud-init status indicates the use of a fallback
       datasource due to the failure of IPv6 connectivity checks.
    3. Ensures that the status message confirms the failure to fetch IMDS
       metadata from the IPv6 address and instead used the fallback datasource.
    4. The cloud-init status exit code should indicate a failure.
    Args:
        client (IntegrationInstance): The integration test client instance.
        address (str): The IPv6 address to check in the cloud-init log.
    """
    log = client.read_from_file("/var/log/cloud-init.log")
    matching_line = None
    for line in log.splitlines():
        if _find_imds_read_log(line, IPV6_METADATA_ADDRESS):
            matching_line = line
            break

    # ensure that the line was NOT found
    assert matching_line is None

    # now check cloud-init status to ensure that ipv6 instance failed to
    # retrieve metadata from IMDS
    status = client.execute("cloud-init status --long")
    assert "Used fallback datasource" in status
    # since IPV6 connectivity checks will fail in Oracle Datasource,
    # we want to verify that it did NOT try querying the IPV6 address
    # and instead used the fallback datasource
    assert (
        "Failed to fetch IMDS metadata from any of: "
        f"http://{IPV4_METADATA_ADDRESS}" in status
    )
    assert not status.ok


def _install_netfilter_perstistent(client: IntegrationInstance):
    assert client.execute("sudo apt-get install -y iptables-persistent").ok


def _clean_and_wait_for_cloudinit(client: IntegrationInstance):
    """
    Clean up the instance and wait for cloud-init to finish

    This function performs the following steps:
    1. Cleans up the instance and wipes all cloud-init logs and data, and then
         reboots the instance.
    2. Waits for the instance to reboot and for cloud-init to finish running.
    """
    client.execute("cloud-init clean --logs")
    client.restart()
    # then wait on cloud-init to finish
    client.instance._wait_for_execute()


@pytest.mark.unstable
@pytest.mark.skipif(PLATFORM != "oci", reason="test is oci specific")
def test_single_stack(client: IntegrationInstance):
    """
    Test the behavior of cloud-init when interacting with the Oracle Cloud
    Infrastructure (OCI) Instance Metadata Service (IMDS) over IPv6, while
    selectively blocking IPv4 and IPv6 traffic.

    This test performs the following steps:
    1. Ensures the IPv6 IMDS is reachable via curl.
    2. Blocks IPv4 traffic and verifies that the IPv6 IMDS is still reachable.
    3. Blocks IPv6 traffic and verifies that the IPv6 IMDS is no longer
       reachable.
    4. Cleans up iptables rules to restore normal network behavior.

    The test is marked as unstable because it requires a specially configured
    instance that is running a private custom-made IPv6-only image that is not
    generally available. Thus, this test is only meant to be run manually by
    developers who have access to the necessary resources.
    """
    _install_netfilter_perstistent(client)

    # Ensure IPv6 is not disabled from a previous test
    client.execute(f"ip6tables -D OUTPUT -d {IPV6_METADATA_ADDRESS} -j REJECT")
    client.execute(f"ip6tables -D INPUT -s {IPV6_METADATA_ADDRESS} -j DROP")
    # Ensure IPv4 is not disabled from a previous test
    client.execute(f"iptables -D OUTPUT -d {IPV4_METADATA_ADDRESS} -j REJECT")
    client.execute(f"iptables -D INPUT -s {IPV4_METADATA_ADDRESS} -j DROP")

    # assert ipv6 imds is reachable via curl
    assert client.execute(
        "curl -f -g -6 -L http://[fd00:c1::a9fe:a9fe]/opc/v1/instance/"
    ).ok

    # Drop IPv4 responses
    assert client.execute(
        f"iptables -I INPUT -s {IPV4_METADATA_ADDRESS} -j DROP"
    ).ok
    # Block IPv4 requests
    assert client.execute(
        f"iptables -I OUTPUT -d {IPV4_METADATA_ADDRESS} -j REJECT"
    ).ok
    # Save the rules so they persist across reboot
    assert client.execute("sudo netfilter-persistent save").ok

    # assert ipv6 imds is reachable via curl when ipv4 is blocked
    assert client.execute(
        "curl -f -g -6 -L http://[fd00:c1::a9fe:a9fe]/opc/v1/instance/"
    ).ok

    _clean_and_wait_for_cloudinit(client)
    # cloud-init should be able to use ipv6 IMDS when ipv4 is blocked
    _test_reading_ipv6_metadata_succeeds(client)

    # Drop IPv6 responses
    assert client.execute(
        f"ip6tables -I INPUT -s {IPV6_METADATA_ADDRESS} -j DROP"
    ).ok
    # Block IPv6 requests
    assert client.execute(
        f"ip6tables -I OUTPUT -d {IPV6_METADATA_ADDRESS} -j REJECT"
    ).ok
    # Save the rules so they persist across reboot
    assert client.execute("sudo netfilter-persistent save").ok
    # assert that curling the ipv6 address fails when ipv6 is blocked
    assert not client.execute(
        "curl -f -g -6 -L http://[fd00:c1::a9fe:a9fe]/opc/v1/instance/"
    ).ok

    _clean_and_wait_for_cloudinit(client)
    # cloud-init should NOT be able to use ipv6 IMDS when ipv6 is blocked
    _test_reading_ipv6_metadata_fails(client)

    # Re-enable IPv6
    client.execute(f"ip6tables -D OUTPUT -d {IPV6_METADATA_ADDRESS} -j REJECT")
    client.execute(f"ip6tables -D INPUT -s {IPV6_METADATA_ADDRESS} -j DROP")
    # Re-enable IPv4
    client.execute(f"iptables -D OUTPUT -d {IPV4_METADATA_ADDRESS} -j REJECT")
    client.execute(f"iptables -D INPUT -s {IPV4_METADATA_ADDRESS} -j DROP")



# This file is part of cloud-init. See LICENSE file for license information.
import re

import pytest

from cloudinit.util import is_true
from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.util import (
    get_feature_flag_value,
    network_wait_logged,
    verify_clean_boot,
    verify_clean_log,
)

USER_DATA = """\
## template: jinja
#cloud-boothook
#!/bin/sh
# Error below will generate stderr
BOOTHOOK/0
echo BOOTHOOKstdout
echo "BOOTHOOK: {{ v1.instance_id }}: is called every boot." >> /boothook.txt
"""


@pytest.mark.user_data(USER_DATA)
class TestBoothook:
    def test_boothook_header_runs_part_per_instance(
        self,
        class_client: IntegrationInstance,
    ):
        """Test boothook handling creates a script that runs per-boot.
        Streams stderr and stdout are directed to
        /var/log/cloud-init-output.log.
        """
        client = class_client
        instance_id = client.instance.execute("cloud-init query instance-id")
        RE_BOOTHOOK = f"BOOTHOOK: {instance_id}: is called every boot"
        log = client.read_from_file("/var/log/cloud-init.log")
        verify_clean_log(log)
        verify_clean_boot(client)
        output = client.read_from_file("/boothook.txt")
        assert 1 == len(re.findall(RE_BOOTHOOK, output))
        client.restart()
        output = client.read_from_file("/boothook.txt")
        assert 2 == len(re.findall(RE_BOOTHOOK, output))
        output_log = client.read_from_file("/var/log/cloud-init-output.log")
        expected_msgs = [
            "BOOTHOOKstdout",
            "boothooks/part-001: 3: BOOTHOOK/0: not found",
        ]
        for msg in expected_msgs:
            assert msg in output_log

    def test_boothook_waits_for_network(
        self, class_client: IntegrationInstance
    ):
        """Test boothook handling waits for network before running."""
        client = class_client
        assert network_wait_logged(
            client.read_from_file("/var/log/cloud-init.log")
        ) == is_true(get_feature_flag_value(client, "MANUAL_NETWORK_WAIT"))



# Author: Jonas Keidel <jonas.keidel@hetzner.com>
# Author: Markus Schade <markus.schade@hetzner.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
#
"""Hetzner Cloud API Documentation
https://docs.hetzner.cloud/"""

import logging

import cloudinit.sources.helpers.hetzner as hc_helper
from cloudinit import dmi, net, sources, url_helper, util
from cloudinit.event import EventScope, EventType
from cloudinit.net.dhcp import NoDHCPLeaseError
from cloudinit.net.ephemeral import EphemeralIPNetwork

LOG = logging.getLogger(__name__)

BUILTIN_DS_CONFIG = {
    "metadata_path": "metadata",
    "metadata_private_networks_path": "metadata/private-networks",
    "userdata_path": "userdata",
}

MD_RETRIES = 60
MD_TIMEOUT = 2
MD_WAIT_RETRY = 2
MD_MAX_WAIT = 120
MD_SLEEP_TIME = 2

# Do not re-configure the network on non-Hetzner network interface
# changes. Currently, Hetzner private network addresses start with 0x86.
EXTRA_HOTPLUG_UDEV_RULES = """
SUBSYSTEM=="net", ATTR{address}=="86:*", GOTO="cloudinit_hook"
GOTO="cloudinit_end"
"""


def base_urls_v1():
    return (
        f"http://[fe80::a9fe:a9fe%25{net.find_fallback_nic()}]/hetzner/v1/",
        "http://169.254.169.254/hetzner/v1/",
    )


class DataSourceHetzner(sources.DataSource):
    dsname = "Hetzner"

    default_update_events = {
        EventScope.NETWORK: {
            EventType.BOOT_NEW_INSTANCE,
            EventType.BOOT,
            EventType.HOTPLUG,
        }
    }

    def __init__(self, sys_cfg, distro, paths):
        sources.DataSource.__init__(self, sys_cfg, distro, paths)
        self.distro = distro
        self.metadata = {}
        self.ds_cfg = util.mergemanydict(
            [
                util.get_cfg_by_path(sys_cfg, ["datasource", "Hetzner"], {}),
                BUILTIN_DS_CONFIG,
            ]
        )
        self.metadata_path = self.ds_cfg["metadata_path"]
        self.metadata_private_networks_path = self.ds_cfg[
            "metadata_private_networks_path"
        ]
        self.userdata_path = self.ds_cfg["userdata_path"]
        self.retries = self.ds_cfg.get("retries", MD_RETRIES)
        self.timeout = self.ds_cfg.get("timeout", MD_TIMEOUT)
        self.wait_retry = self.ds_cfg.get("wait_retry", MD_WAIT_RETRY)
        self.max_wait = self.ds_cfg.get("max_wait", MD_MAX_WAIT)
        self.sleep_time = self.ds_cfg.get("sleep_time", MD_SLEEP_TIME)
        self._network_config = sources.UNSET
        self.dsmode = sources.DSMODE_NETWORK
        self.metadata_full = None

        self.extra_hotplug_udev_rules = EXTRA_HOTPLUG_UDEV_RULES

    def _unpickle(self, ci_pkl_version: int) -> None:
        super()._unpickle(ci_pkl_version)
        self.extra_hotplug_udev_rules = EXTRA_HOTPLUG_UDEV_RULES
        self.wait_retry = self.ds_cfg.get("wait_retry", MD_WAIT_RETRY)
        self.max_wait = self.ds_cfg.get("max_wait", MD_MAX_WAIT)
        self.sleep_time = self.ds_cfg.get("sleep_time", MD_SLEEP_TIME)
        self.metadata_path = self.ds_cfg["metadata_path"]
        self.metadata_private_networks_path = self.ds_cfg[
            "metadata_private_networks_path"
        ]
        self.userdata_path = self.ds_cfg["userdata_path"]

    def _get_data(self):
        (on_hetzner, serial) = get_hcloud_data()

        if not on_hetzner:
            return False

        base_urls = base_urls_v1()
        try:
            with EphemeralIPNetwork(
                self.distro,
                interface=net.find_fallback_nic(),
                ipv4=True,
                ipv6=True,
                connectivity_urls_data=[
                    {
                        "url": url_helper.combine_url(
                            url, f"{self.metadata_path}/instance-id"
                        )
                    }
                    for url in base_urls
                ],
            ):
                url, contents = hc_helper.get_metadata(
                    [
                        url_helper.combine_url(url, self.metadata_path)
                        for url in base_urls
                    ],
                    max_wait=self.max_wait,
                    timeout=self.timeout,
                    sleep_time=self.sleep_time,
                )
                LOG.debug("Using metadata source: '%s'", url)
                md = util.load_yaml(contents.decode(), allowed=(dict, list))
                url, contents = hc_helper.get_metadata(
                    [
                        url_helper.combine_url(
                            url, self.metadata_private_networks_path
                        )
                        for url in base_urls
                    ],
                    max_wait=self.max_wait,
                    timeout=self.timeout,
                    sleep_time=self.sleep_time,
                )
                LOG.debug("Using private_networks source: '%s'", url)
                md["private-networks"] = util.load_yaml(
                    contents.decode(), allowed=(dict, list)
                )
                url, ud = hc_helper.get_metadata(
                    [
                        url_helper.combine_url(url, self.userdata_path)
                        for url in base_urls
                    ],
                    max_wait=self.max_wait,
                    timeout=self.timeout,
                    sleep_time=self.sleep_time,
                )
                LOG.debug("Using userdata source: '%s'", url)
                if not ud:
                    LOG.debug("Got empty userdata")
        except NoDHCPLeaseError as e:
            LOG.error("Bailing, DHCP Exception: %s", e)
            raise

        # Hetzner cloud does not support binary user-data. So here, do a
        # base64 decode of the data if we can. The end result being that a
        # user can provide base64 encoded (possibly gzipped) data as user-data.
        #
        # The fallout is that in the event of b64 encoded user-data,
        # /var/lib/cloud-init/cloud-config.txt will not be identical to the
        # user-data provided.  It will be decoded.
        self.userdata_raw = util.maybe_b64decode(ud)
        self.metadata_full = md

        # hostname is name provided by user at launch.  The API enforces it is
        # a valid hostname, but it is not guaranteed to be resolvable in dns or
        # fully qualified.
        self.metadata["instance-id"] = md["instance-id"]
        self.metadata["local-hostname"] = md["hostname"]
        self.metadata["network-config"] = md.get("network-config", None)
        self.metadata["public-keys"] = md.get("public-keys", None)
        self.metadata["private-networks"] = md.get("private-networks", [])
        self.vendordata_raw = md.get("vendor_data", None)

        # instance-id and serial from SMBIOS should be identical
        if self.get_instance_id() != serial:
            raise RuntimeError(
                "SMBIOS serial does not match instance ID from metadata"
            )

        return True

    def check_instance_id(self, sys_cfg):
        return sources.instance_id_matches_system_uuid(
            self.get_instance_id(), "system-serial-number"
        )

    @property
    def network_config(self):
        """Configure the networking. This needs to be done each boot, since
        the IP information may have changed due to snapshot and/or
        migration.
        """

        if self._network_config is None:
            LOG.warning(
                "Found None as cached _network_config. Resetting to %s",
                sources.UNSET,
            )
            self._network_config = sources.UNSET

        if self._network_config != sources.UNSET:
            return self._network_config

        net_config = self.metadata["network-config"]
        if not net_config:
            raise RuntimeError("Unable to get meta-data from server....")

        private_networks = self.metadata.get("private-networks", [])
        private_networks_config = []
        for private_network in private_networks:
            private_networks_config.append(
                {
                    "type": "physical",
                    "mac_address": private_network["mac_address"],
                    "name": hc_helper.get_interface_name_from_mac(
                        private_network["mac_address"]
                    ),
                    "subnets": [
                        {
                            "ipv4": True,
                            "type": "dhcp",
                        }
                    ],
                }
            )
        net_config["config"].extend(private_networks_config)
        self._network_config = net_config
        return self._network_config


def get_hcloud_data():
    vendor_name = dmi.read_dmi_data("system-manufacturer")
    if vendor_name != "Hetzner":
        return False, None

    serial = dmi.read_dmi_data("system-serial-number")
    if serial:
        LOG.debug("Running on Hetzner Cloud: serial=%s", serial)
    else:
        raise RuntimeError("Hetzner Cloud detected, but no serial found")

    return True, serial


# Used to match classes to dependencies
datasources = [
    (DataSourceHetzner, (sources.DEP_FILESYSTEM,)),
]


# Return a list of data sources that match this set of dependencies
def get_datasource_list(depends):
    return sources.list_from_depends(depends, datasources)



import logging
import os
import sys

from cloudinit.performance import timed

LOG = logging.getLogger(__name__)


def logexc(
    log, msg, *args, log_level: int = logging.WARNING, exc_info=True
) -> None:
    log.log(log_level, msg, *args)
    log.debug(msg, exc_info=exc_info, *args)


@timed("Writing to console")
def write_to_console(conpath, text):
    with open(conpath, "w") as wfh:
        wfh.write(text)
        wfh.flush()


def multi_log(
    text,
    console=True,
    stderr=True,
    log=None,
    log_level=logging.DEBUG,
    fallback_to_stdout=True,
):
    if stderr:
        sys.stderr.write(text)
    if console:
        conpath = "/dev/console"
        writing_to_console_worked = False
        if os.path.exists(conpath):
            try:
                write_to_console(conpath, text)
                writing_to_console_worked = True
            except OSError:
                console_error = "Failed to write to /dev/console"
                sys.stdout.write(f"{console_error}\n")
                if log:
                    log.log(logging.WARNING, console_error)

        if fallback_to_stdout and not writing_to_console_worked:
            # A container may lack /dev/console (arguably a container bug).
            # Additionally, /dev/console may not be writable to on a VM (again
            # likely a VM bug or virtualization bug).
            #
            # If either of these is the case, then write output to stdout.
            # This will result in duplicate stderr and stdout messages if
            # stderr was True.
            #
            # even though systemd might have set up output to go to
            # /dev/console, the user may have configured elsewhere via
            # cloud-config 'output'.  If there is /dev/console, messages will
            # still get there.
            sys.stdout.write(text)
    if log:
        if text[-1] == "\n":
            log.log(log_level, text[:-1])
        else:
            log.log(log_level, text)


def error(msg, rc=1, fmt="Error:\n{}", sys_exit=False):
    r"""Print error to stderr and return or exit

    @param msg: message to print
    @param rc: return code (default: 1)
    @param fmt: format string for putting message in (default: 'Error:\n {}')
    @param sys_exit: exit when called (default: false)
    """
    print(fmt.format(msg), file=sys.stderr)
    if sys_exit:
        sys.exit(rc)
    return rc



# This file is part of cloud-init. See LICENSE file for license information.

import logging
from unittest import mock

import pytest

from cloudinit import util
from cloudinit.config.cc_write_files_deferred import handle
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import skipUnlessJsonSchema
from tests.unittests.util import get_cloud

LOG = logging.getLogger(__name__)


@pytest.mark.usefixtures("fake_filesystem")
class TestWriteFilesDeferred:

    USER = "root"

    @mock.patch("cloudinit.config.cc_write_files.util.chownbyname")
    def test_filtering_deferred_files(self, m_chownbyname):
        expected = "hello world\n"
        config = {
            "write_files": [
                {
                    "path": "/tmp/deferred.file",
                    "defer": True,
                    "content": expected,
                },
                {"path": "/tmp/not_deferred.file"},
            ]
        }
        cc = get_cloud("ubuntu")
        # fake_filesytem's tree is owned by $USER:$USER
        with mock.patch.object(
            cc.distro, "default_owner", f"{self.USER}:{self.USER}"
        ):
            handle("cc_write_files_deferred", config, cc, [])
        assert util.load_text_file("/tmp/deferred.file") == expected
        with pytest.raises(FileNotFoundError):
            util.load_text_file("/tmp/not_deferred.file")
        assert [
            mock.call(mock.ANY, self.USER, self.USER)
        ] == m_chownbyname.call_args_list


class TestWriteFilesDeferredSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        [
            # Allow undocumented keys client keys without error
            (
                {"write_files": [{"defer": "no"}]},
                "write_files.0.defer: 'no' is not of type 'boolean'",
            ),
        ],
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        with pytest.raises(SchemaValidationError, match=error_msg):
            validate_cloudconfig_schema(config, get_schema(), strict=True)



# Copyright (C) 2012 Yahoo! Inc.
#
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

try:
    from cloudinit import meson_versions  # type: ignore

    __VERSION__ = meson_versions.UPSTREAM_VERSION
    _DOWNSTREAM_VERSION = meson_versions.DOWNSTREAM_VERSION
except ImportError:
    __VERSION__ = "@MISSING_MESON_BUILD_ARTIFACT@"
    _DOWNSTREAM_VERSION = "@DOWNSTREAM_VERSION@"  # Optional for packagers


FEATURES = [
    # supports network config version 1
    "NETWORK_CONFIG_V1",
    # supports network config version 2 (netplan)
    "NETWORK_CONFIG_V2",
]


def version_string():
    """Extract a version string from cloud-init."""
    if "DOWNSTREAM_VERSION" not in _DOWNSTREAM_VERSION:
        return _DOWNSTREAM_VERSION
    return __VERSION__



# This file is part of cloud-init. See LICENSE file for license information.

"""Autoinstall: Support ubuntu live-server autoinstall syntax."""

import logging
import re

from cloudinit import subp, util
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema
from cloudinit.settings import PER_ONCE

LOG = logging.getLogger(__name__)

meta: MetaSchema = {
    "id": "cc_ubuntu_autoinstall",
    "distros": ["ubuntu"],
    "frequency": PER_ONCE,
    "activate_by_schema_keys": ["autoinstall"],
}


LIVE_INSTALLER_SNAPS = ("subiquity", "ubuntu-desktop-installer")


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:

    util.wait_for_snap_seeded(cloud)
    snap_list, _ = subp.subp(["snap", "list"])
    installer_present = None
    for snap_name in LIVE_INSTALLER_SNAPS:
        if re.search(snap_name, snap_list):
            installer_present = snap_name
    if not installer_present:
        LOG.warning(
            "Skipping autoinstall module. Expected one of the Ubuntu"
            " installer snap packages to be present: %s",
            ", ".join(LIVE_INSTALLER_SNAPS),
        )
        return
    LOG.debug(
        "Valid autoinstall schema. Config will be processed by %s",
        installer_present,
    )



# Copyright (C) 2015 Canonical Ltd.
# Copyright (C) 2017-2019 VMware INC.
#
# Author: Maitreyee Saikia <msaikia@vmware.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import os
import stat
from unittest import mock

import pytest

from cloudinit import util
from cloudinit.sources.helpers.vmware.imc.config_custom_script import (
    CustomScriptConstant,
    CustomScriptNotFound,
    PostCustomScript,
    PreCustomScript,
)


@pytest.fixture
def fake_exec_dir(mocker, tmp_path):
    exec_dir = tmp_path / ".customization"
    mocker.patch.object(CustomScriptConstant, "CUSTOM_TMP_DIR", str(exec_dir))
    return exec_dir


@pytest.fixture
def fake_exec_script(fake_exec_dir, mocker):
    ex_script = fake_exec_dir / ".customize.sh"
    mocker.patch.object(CustomScriptConstant, "CUSTOM_SCRIPT", str(ex_script))
    return ex_script


class TestVmwareCustomScript:
    def test_prepare_custom_script(self, fake_exec_script, tmp_path):
        """
        This test is designed to verify the behavior based on the presence of
        custom script. Mainly needed for scenario where a custom script is
        expected, but was not properly copied. "CustomScriptNotFound" exception
        is raised in such cases.
        """
        # Custom script does not exist.
        preCust = PreCustomScript("random-vmw-test", str(tmp_path))
        assert "random-vmw-test" == preCust.scriptname
        assert str(tmp_path) == preCust.directory
        assert str(tmp_path / "random-vmw-test") == preCust.scriptpath
        with pytest.raises(CustomScriptNotFound):
            preCust.prepare_script()

        # Custom script exists.
        custScript = str(tmp_path / "test-cust")
        util.write_file(custScript, "test-CR-strip\r\r")

        postCust = PostCustomScript("test-cust", str(tmp_path), str(tmp_path))
        assert "test-cust" == postCust.scriptname
        assert str(tmp_path) == preCust.directory
        assert custScript == postCust.scriptpath
        postCust.prepare_script()

        # Custom script is copied with exec privilege
        assert fake_exec_script.exists()
        st = os.stat(fake_exec_script)
        assert st.st_mode & stat.S_IEXEC
        assert "test-CR-strip" == fake_exec_script.read_text()

    def test_execute_post_cust(self, fake_exec_script, tmp_path):
        """
        This test is designed to verify the behavior after execute post
        customization.
        """
        # Prepare the customize package
        postCustRun = str(tmp_path / "post-customize-guest.sh")
        util.write_file(postCustRun, "This is the script to run post cust")
        userScript = str(tmp_path / "test-cust")
        util.write_file(userScript, "This is the post cust script")

        # Mock the cc_scripts_per_instance dir and marker file.
        # Create another tmp dir for cc_scripts_per_instance.
        ccScriptDir = tmp_path / "out"
        ccScriptDir.mkdir()
        ccScript = ccScriptDir / "post-customize-guest.sh"
        markerFile = tmp_path / ".markerFile"

        with mock.patch.object(
            CustomScriptConstant,
            "POST_CUSTOM_PENDING_MARKER",
            str(markerFile),
        ):
            postCust = PostCustomScript(
                "test-cust", str(tmp_path), ccScriptDir
            )
            postCust.execute()
            # Check cc_scripts_per_instance and marker file
            # are created.
            assert ccScript.exists()
            assert (
                "This is the script to run post cust" == ccScript.read_text()
            )
            assert markerFile.exists()



# This file is part of cloud-init. See LICENSE file for license information.

import os
from unittest import mock

import pytest

from cloudinit import distros, util
from cloudinit.distros.ubuntu import Distro
from cloudinit.net.dhcp import Dhcpcd, IscDhclient, Udhcpc

M_PATH = "cloudinit.distros."

unknown_arch_info = {
    "arches": ["default"],
    "failsafe": {
        "primary": "http://fs-primary-default",
        "security": "http://fs-security-default",
    },
}

package_mirrors = [
    {
        "arches": ["i386", "amd64"],
        "failsafe": {
            "primary": "http://fs-primary-intel",
            "security": "http://fs-security-intel",
        },
        "search": {
            "primary": [
                "http://%(ec2_region)s.ec2/",
                "http://%(availability_zone)s.clouds/",
            ],
            "security": [
                "http://security-mirror1-intel",
                "http://security-mirror2-intel",
            ],
        },
    },
    {
        "arches": ["armhf", "armel"],
        "failsafe": {
            "primary": "http://fs-primary-arm",
            "security": "http://fs-security-arm",
        },
    },
    unknown_arch_info,
]

gpmi = distros._get_package_mirror_info
gapmi = distros._get_arch_package_mirror_info


@pytest.mark.usefixtures("fake_filesystem")
class TestGenericDistro:
    def _write_load_doas(self, user, rules):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        if not os.path.exists("/etc"):
            os.makedirs("/etc")
        d.write_doas_rules(user, rules)
        contents = util.load_text_file(d.doas_fn)
        return contents, cls, d

    def _write_load_sudoers(self, _user, rules):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        os.makedirs("/etc")
        os.makedirs(os.path.join("/etc", "sudoers.d"))
        d.write_sudo_rules("harlowja", rules)
        contents = util.load_text_file(d.ci_sudoers_fn)
        return contents, cls, d

    def _count_in(self, lines_look_for, text_content):
        found_amount = 0
        for e in lines_look_for:
            for line in text_content.splitlines():
                line = line.strip()
                if line == e:
                    found_amount += 1
        return found_amount

    def test_doas_ensure_rules(self):
        rules = ["permit nopass harlowja"]
        contents = self._write_load_doas("harlowja", rules)[0]
        expected = ["permit nopass harlowja"]
        assert len(expected) == self._count_in(expected, contents)

    def test_doas_ensure_rules_list(self):
        rules = [
            "permit nopass harlowja cmd ls",
            "permit nopass harlowja cmd pwd",
            "permit nopass harlowja cmd df",
        ]
        contents = self._write_load_doas("harlowja", rules)[0]
        expected = [
            "permit nopass harlowja cmd ls",
            "permit nopass harlowja cmd pwd",
            "permit nopass harlowja cmd df",
        ]
        assert len(expected) == self._count_in(expected, contents)

    def test_doas_ensure_handle_duplicates(self):
        rules = [
            "permit nopass harlowja cmd ls",
            "permit nopass harlowja cmd pwd",
            "permit nopass harlowja cmd df",
        ]
        d = self._write_load_doas("harlowja", rules)[2]
        # write to doas.conf again - should not create duplicate rules
        d.write_doas_rules("harlowja", rules)
        contents = util.load_text_file(d.doas_fn)
        expected = [
            "permit nopass harlowja cmd ls",
            "permit nopass harlowja cmd pwd",
            "permit nopass harlowja cmd df",
        ]
        assert len(expected) == self._count_in(expected, contents)

    def test_doas_ensure_new(self):
        rules = [
            "permit nopass harlowja cmd ls",
            "permit nopass harlowja cmd pwd",
            "permit nopass harlowja cmd df",
        ]
        contents = self._write_load_doas("harlowja", rules)[0]
        assert "# Created by cloud-init v." in contents
        assert "harlowja" in contents
        assert 4 == contents.count("harlowja")

    def test_doas_ensure_append(self):
        util.write_file("/etc/doas.conf", "# root user\npermit nopass root\n")
        rules = [
            "permit nopass harlowja cmd ls",
            "permit nopass harlowja cmd pwd",
            "permit nopass harlowja cmd df",
        ]
        contents = self._write_load_doas("harlowja", rules)[0]
        assert "root" in contents
        assert 2 == contents.count("root")
        assert "harlowja" in contents
        assert 4 == contents.count("harlowja")

    def test_sudoers_ensure_rules(self):
        rules = "ALL=(ALL:ALL) ALL"
        contents = self._write_load_sudoers("harlowja", rules)[0]
        expected = ["harlowja ALL=(ALL:ALL) ALL"]
        assert len(expected) == self._count_in(expected, contents)
        not_expected = [
            "harlowja A",
            "harlowja L",
            "harlowja L",
        ]
        assert 0 == self._count_in(not_expected, contents)

    def test_sudoers_ensure_rules_list(self):
        rules = [
            "ALL=(ALL:ALL) ALL",
            "B-ALL=(ALL:ALL) ALL",
            "C-ALL=(ALL:ALL) ALL",
        ]
        contents = self._write_load_sudoers("harlowja", rules)[0]
        expected = [
            "harlowja ALL=(ALL:ALL) ALL",
            "harlowja B-ALL=(ALL:ALL) ALL",
            "harlowja C-ALL=(ALL:ALL) ALL",
        ]
        assert len(expected) == self._count_in(expected, contents)
        not_expected = [
            "harlowja A",
            "harlowja L",
            "harlowja L",
        ]
        assert 0 == self._count_in(not_expected, contents)

    def test_sudoers_ensure_handle_duplicates(self):
        rules = [
            "ALL=(ALL:ALL) ALL",
            "B-ALL=(ALL:ALL) ALL",
            "C-ALL=(ALL:ALL) ALL",
        ]
        d = self._write_load_sudoers("harlowja", rules)[2]
        # write to sudoers again - should not create duplicate rules
        d.write_sudo_rules("harlowja", rules)
        contents = util.load_text_file(d.ci_sudoers_fn)
        expected = [
            "harlowja ALL=(ALL:ALL) ALL",
            "harlowja B-ALL=(ALL:ALL) ALL",
            "harlowja C-ALL=(ALL:ALL) ALL",
        ]
        assert len(expected) == self._count_in(expected, contents)
        not_expected = [
            "harlowja A",
            "harlowja L",
            "harlowja L",
        ]
        assert 0 == self._count_in(not_expected, contents)

    def test_sudoers_ensure_new(self):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        d.ensure_sudo_dir("/b")
        contents = util.load_text_file("/etc/sudoers")
        assert "includedir /b" in contents
        assert os.path.isdir("/b")

    def test_sudoers_ensure_append(self, caplog):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        util.write_file("/etc/sudoers", "josh, josh\n")
        d.ensure_sudo_dir("/b")
        contents = util.load_text_file("/etc/sudoers")
        assert "includedir /b" in contents
        assert os.path.isdir("/b")
        assert "josh" in contents
        assert 2 == contents.count("josh")
        assert "Added '#includedir /b' to /etc/sudoers" in caplog.text

    def test_sudoers_ensure_append_sudoer_file(self):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        util.write_file("/etc/sudoers", "josh, josh\n")
        d.ensure_sudo_dir("/b", "/etc/sudoers")
        contents = util.load_text_file("/etc/sudoers")
        assert "includedir /b" in contents
        assert os.path.isdir("/b")
        assert "josh" in contents
        assert 2 == contents.count("josh")

    def test_usr_sudoers_ensure_new(self, caplog):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        util.write_file("/usr/etc/sudoers", "josh, josh\n")
        d.ensure_sudo_dir("/b")
        contents = util.load_text_file("/etc/sudoers")
        assert "josh" in contents
        assert 2 == contents.count("josh")
        assert "includedir /b" in contents
        assert os.path.isdir("/b")
        assert "Using content from '/usr/etc/sudoers" in caplog.text

    def test_usr_sudoers_ensure_no_etc_create_when_include_in_usr_etc(self):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        util.write_file("/usr/etc/sudoers", "#includedir /b")
        d.ensure_sudo_dir("/b")
        assert not os.path.exists("/etc/sudoers")

    def test_sudoers_ensure_only_one_includedir(self):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        for char in ["#", "@"]:
            util.write_file("/etc/sudoers", "{}includedir /b".format(char))
            d.ensure_sudo_dir("/b")
            contents = util.load_text_file("/etc/sudoers")
            assert "includedir /b" in contents
            assert os.path.isdir("/b")
            assert 1 == contents.count("includedir /b")

    def test_arch_package_mirror_info_unknown(self):
        """for an unknown arch, we should get back that with arch 'default'."""
        arch_mirrors = gapmi(package_mirrors, arch="unknown")
        assert unknown_arch_info == arch_mirrors

    def test_arch_package_mirror_info_known(self):
        arch_mirrors = gapmi(package_mirrors, arch="amd64")
        assert package_mirrors[0] == arch_mirrors

    def test_systemd_in_use(self):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        os.makedirs("/run/systemd/system")
        assert d.uses_systemd()

    def test_systemd_not_in_use(self):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        assert not d.uses_systemd()

    def test_systemd_symlink(self):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        os.makedirs("/run/systemd")
        os.symlink("/", "/run/systemd/system")
        assert not d.uses_systemd()

    @mock.patch("cloudinit.distros.debian.read_system_locale")
    def test_get_locale_ubuntu(self, m_locale):
        """Test ubuntu distro returns locale set to C.UTF-8"""
        m_locale.return_value = "C.UTF-8"
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        locale = d.get_locale()
        assert "C.UTF-8" == locale

    @mock.patch("cloudinit.distros.rhel.Distro._read_system_locale")
    def test_get_locale_rhel(self, m_locale):
        """Test rhel distro returns locale set to C.UTF-8"""
        m_locale.return_value = "C.UTF-8"
        cls = distros.fetch("rhel")
        d = cls("rhel", {}, None)
        locale = d.get_locale()
        assert "C.UTF-8" == locale

    def test_expire_passwd_uses_chpasswd(self):
        """Test ubuntu.expire_passwd uses the passwd command."""
        for d_name in ("ubuntu", "rhel"):
            cls = distros.fetch(d_name)
            d = cls(d_name, {}, None)
            with mock.patch("cloudinit.subp.subp") as m_subp:
                d.expire_passwd("myuser")
            m_subp.assert_called_once_with(["passwd", "--expire", "myuser"])

    def test_expire_passwd_freebsd_uses_pw_command(self):
        """Test FreeBSD.expire_passwd uses the pw command."""
        cls = distros.fetch("freebsd")
        # patch ifconfig -a
        with mock.patch(
            "cloudinit.distros.networking.subp.subp", return_value=("", None)
        ):
            d = cls("freebsd", {}, None)
        with mock.patch("cloudinit.subp.subp") as m_subp:
            d.expire_passwd("myuser")
        m_subp.assert_called_once_with(
            ["pw", "usermod", "myuser", "-p", "01-Jan-1970"]
        )

    @mock.patch(M_PATH + "subp.subp")
    def test_do_as_nested_py_cmd(self, m_subp):
        cls = distros.fetch("ubuntu")
        d = cls("ubuntu", {}, None)
        d.do_as(
            command=[
                "/usr/bin/python3",
                "-c",
                "import site; print(site.getuserbase())",
            ],
            user="ubuntu",
        )
        m_subp.assert_called_once_with(
            [
                "su",
                "-",
                "ubuntu",
                "-c",
                "env PATH=$PATH /usr/bin/python3 -c 'import site; print(site.getuserbase())'",  # noqa: E501
            ]
        )


class TestGetPackageMirrors:
    def return_first(self, mlist):
        if not mlist:
            return None
        return mlist[0]

    def return_second(self, mlist):
        if not mlist:
            return None

        return mlist[1] if len(mlist) > 1 else None

    def return_none(self, _mlist):
        return None

    def return_last(self, mlist):
        if not mlist:
            return None
        return mlist[-1]

    @pytest.mark.parametrize(
        "allow_ec2_mirror, platform_type, mirrors",
        [
            (
                True,
                "ec2",
                [
                    {
                        "primary": "http://us-east-1.ec2/",
                        "security": "http://security-mirror1-intel",
                    },
                    {
                        "primary": "http://us-east-1a.clouds/",
                        "security": "http://security-mirror2-intel",
                    },
                ],
            ),
            (
                True,
                "other",
                [
                    {
                        "primary": "http://us-east-1.ec2/",
                        "security": "http://security-mirror1-intel",
                    },
                    {
                        "primary": "http://us-east-1a.clouds/",
                        "security": "http://security-mirror2-intel",
                    },
                ],
            ),
            (
                False,
                "ec2",
                [
                    {
                        "primary": "http://us-east-1.ec2/",
                        "security": "http://security-mirror1-intel",
                    },
                    {
                        "primary": "http://us-east-1a.clouds/",
                        "security": "http://security-mirror2-intel",
                    },
                ],
            ),
            (
                False,
                "other",
                [
                    {
                        "primary": "http://us-east-1a.clouds/",
                        "security": "http://security-mirror1-intel",
                    },
                    {
                        "primary": "http://fs-primary-intel",
                        "security": "http://security-mirror2-intel",
                    },
                ],
            ),
        ],
    )
    def test_get_package_mirror_info_az_ec2(
        self, allow_ec2_mirror, platform_type, mirrors
    ):
        flag_path = (
            "cloudinit.distros.ALLOW_EC2_MIRRORS_ON_NON_AWS_INSTANCE_TYPES"
        )
        with mock.patch(flag_path, allow_ec2_mirror):
            arch_mirrors = gapmi(package_mirrors, arch="amd64")
            data_source_mock = mock.Mock(
                availability_zone="us-east-1a", platform_type=platform_type
            )

            results = gpmi(
                arch_mirrors,
                data_source=data_source_mock,
                mirror_filter=self.return_first,
            )
            assert results == mirrors[0]

            results = gpmi(
                arch_mirrors,
                data_source=data_source_mock,
                mirror_filter=self.return_second,
            )
            assert results == mirrors[1]

            results = gpmi(
                arch_mirrors,
                data_source=data_source_mock,
                mirror_filter=self.return_none,
            )
            assert results == package_mirrors[0]["failsafe"]

    def test_get_package_mirror_info_az_non_ec2(self):
        arch_mirrors = gapmi(package_mirrors, arch="amd64")
        data_source_mock = mock.Mock(availability_zone="nova.cloudvendor")

        results = gpmi(
            arch_mirrors,
            data_source=data_source_mock,
            mirror_filter=self.return_first,
        )
        assert results == {
            "primary": "http://nova.cloudvendor.clouds/",
            "security": "http://security-mirror1-intel",
        }

        results = gpmi(
            arch_mirrors,
            data_source=data_source_mock,
            mirror_filter=self.return_last,
        )
        assert results == {
            "primary": "http://nova.cloudvendor.clouds/",
            "security": "http://security-mirror2-intel",
        }

    def test_get_package_mirror_info_none(self):
        arch_mirrors = gapmi(package_mirrors, arch="amd64")
        data_source_mock = mock.Mock(availability_zone=None)

        # because both search entries here replacement based on
        # availability-zone, the filter will be called with an empty list and
        # failsafe should be taken.
        results = gpmi(
            arch_mirrors,
            data_source=data_source_mock,
            mirror_filter=self.return_first,
        )
        assert results == {
            "primary": "http://fs-primary-intel",
            "security": "http://security-mirror1-intel",
        }

        results = gpmi(
            arch_mirrors,
            data_source=data_source_mock,
            mirror_filter=self.return_last,
        )
        assert results == {
            "primary": "http://fs-primary-intel",
            "security": "http://security-mirror2-intel",
        }


@pytest.mark.usefixtures("fake_filesystem")
class TestDistro:
    @pytest.mark.parametrize("is_noexec", [False, True])
    @mock.patch(M_PATH + "util.has_mount_opt")
    @mock.patch(M_PATH + "temp_utils.get_tmp_ancestor", return_value="/tmp")
    def test_get_tmp_exec_path(
        self, m_get_tmp_ancestor, m_has_mount_opt, is_noexec, mocker
    ):
        m_has_mount_opt.return_value = not is_noexec
        cls = distros.fetch("ubuntu")
        distro = cls("ubuntu", {}, None)
        mocker.patch.object(distro, "usr_lib_exec", "/usr_lib_exec")
        tmp_path = distro.get_tmp_exec_path()
        if is_noexec:
            assert "/tmp" == tmp_path
        else:
            assert "/usr_lib_exec/cloud-init/clouddir" == tmp_path


@pytest.mark.parametrize(
    "chosen_client, config, which_override",
    [
        pytest.param(
            IscDhclient,
            {"network": {"dhcp_client_priority": ["dhclient"]}},
            None,
            id="single_client_is_found_from_config_dhclient",
        ),
        pytest.param(
            Udhcpc,
            {"network": {"dhcp_client_priority": ["udhcpc"]}},
            None,
            id="single_client_is_found_from_config_udhcpc",
        ),
        pytest.param(
            Dhcpcd,
            {"network": {"dhcp_client_priority": ["dhcpcd"]}},
            None,
            id="single_client_is_found_from_config_dhcpcd",
        ),
        pytest.param(
            Dhcpcd,
            {"network": {"dhcp_client_priority": ["dhcpcd", "dhclient"]}},
            None,
            id="first_client_is_found_from_config_dhcpcd",
        ),
        pytest.param(
            Udhcpc,
            {
                "network": {
                    "dhcp_client_priority": ["udhcpc", "dhcpcd", "dhclient"]
                }
            },
            None,
            id="first_client_is_found_from_config_udhcpc",
        ),
        pytest.param(
            Dhcpcd,
            {"network": {"dhcp_client_priority": []}},
            None,
            id="first_client_is_found_no_config_dhcpcd",
        ),
        pytest.param(
            Dhcpcd,
            {
                "network": {
                    "dhcp_client_priority": ["udhcpc", "dhcpcd", "dhclient"]
                }
            },
            [False, False, True, True],
            id="second_client_is_found_from_config_dhcpcd",
        ),
    ],
)
class TestDHCP:
    @mock.patch("cloudinit.net.dhcp.subp.which")
    def test_dhcp_configuration(
        self, m_which, chosen_client, config, which_override
    ):
        """check that, when a user provides a configuration at
        network.dhcp_client_priority, the correct client is chosen
        """
        m_which.side_effect = which_override
        distro = Distro("", {}, {})
        distro._cfg = config
        assert isinstance(distro.dhcp_client, chosen_client)



# Copyright (C) 2018 Jonas Keidel
#
# Author: Jonas Keidel <jonas.keidel@hetzner.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import pytest

from cloudinit import settings, util
from cloudinit.sources import DataSourceHetzner
from tests.unittests.helpers import mock

METADATA = b"""
hostname: cloudinit-test
instance-id: 123456
local-ipv4: ''
network-config:
  config:
  - mac_address: 96:00:00:08:19:da
    name: eth0
    subnets:
    - dns_nameservers:
      - 185.12.64.1
      - 185.12.64.2
      ipv4: true
      type: dhcp
    - address: 2a01:4f8:beef:beef::1/64
      dns_nameservers:
      - 2a01:4ff:ff00::add:2
      - 2a01:4ff:ff00::add:1
      gateway: fe80::1
      ipv6: true
    type: physical
  version: 1
network-sysconfig: "DEVICE='eth0'\nTYPE=Ethernet\nBOOTPROTO=dhcp\n\
  ONBOOT='yes'\nHWADDR=96:00:00:08:19:da\n\
  IPV6INIT=yes\nIPV6ADDR=2a01:4f8:beef:beef::1/64\n\
  IPV6_DEFAULTGW=fe80::1%eth0\nIPV6_AUTOCONF=no\n\
  DNS1=185.12.64.1\nDNS2=185.12.64.2\n"
public-ipv4: 192.168.0.2
public-keys:
- ssh-ed25519 \
  AAAAC3Nzac1lZdI1NTE5AaaAIaFrcac0yVITsmRrmueq6MD0qYNKlEvW8O1Ib4nkhmWh \
  test-key@workstation
vendor_data: "test"
"""

USERDATA = b"""#cloud-config
runcmd:
- [touch, /root/cloud-init-worked ]
"""

PRIVATE_NETWORKS = b"""
- ip: 10.1.0.2
  alias_ips: []
  interface_num: 2
  mac_address: 86:00:00:aa:5d:f8
  network_id: 11352901
  network_name: network-2
  network: 10.1.0.0/16
  subnet: 10.1.0.0/24
  gateway: 10.1.0.1
"""


class TestDataSourceHetzner:
    """
    Test reading the meta-data
    """

    @pytest.fixture
    def ds(self, paths, tmp_path):
        distro = mock.MagicMock()
        distro.get_tmp_exec_path = str(tmp_path)
        ds = DataSourceHetzner.DataSourceHetzner(
            settings.CFG_BUILTIN, distro, paths
        )
        return ds

    @mock.patch("cloudinit.net.dhcp.maybe_perform_dhcp_discovery")
    @mock.patch("cloudinit.sources.DataSourceHetzner.EphemeralIPNetwork")
    @mock.patch("cloudinit.net.find_fallback_nic")
    @mock.patch("cloudinit.sources.helpers.hetzner.get_metadata")
    @mock.patch("cloudinit.sources.DataSourceHetzner.get_hcloud_data")
    def test_read_data(
        self,
        m_get_hcloud_data,
        m_get_metadata,
        m_fallback_nic,
        m_net,
        m_dhcp,
        ds,
    ):
        m_get_hcloud_data.return_value = (
            True,
            str(util.load_yaml(METADATA).get("instance-id")),
        )
        # Use side_effect to return values for the three sequential calls to
        # helpers.hetzner.get_metadata: metadata, private-networks, userdata
        m_get_metadata.side_effect = [
            ("metadata_url", METADATA),
            ("privnets_url", PRIVATE_NETWORKS),
            ("userdata_url", USERDATA),
        ]
        m_fallback_nic.return_value = "eth0"
        m_dhcp.return_value = [
            {
                "interface": "eth0",
                "fixed-address": "192.168.0.2",
                "routers": "192.168.0.1",
                "subnet-mask": "255.255.255.0",
                "broadcast-address": "192.168.0.255",
            }
        ]

        assert True is ds.get_data()

        m_net.assert_called_once_with(
            ds.distro,
            interface="eth0",
            ipv4=True,
            ipv6=True,
            connectivity_urls_data=[
                {
                    "url": "http://[fe80::a9fe:a9fe%25eth0]/hetzner/v1/metadata/instance-id"
                },
                {
                    "url": "http://169.254.169.254/hetzner/v1/metadata/instance-id"
                },
            ],
        )

        assert 0 != m_get_metadata.call_count

        assert (
            util.load_yaml(METADATA).get("hostname")
            == ds.get_hostname().hostname
        )

        assert (
            util.load_yaml(METADATA).get("public-keys")
            == ds.get_public_ssh_keys()
        )

        assert ds.metadata["private-networks"] == util.load_yaml(
            PRIVATE_NETWORKS, allowed=(dict, list)
        )
        assert isinstance(ds.get_public_ssh_keys(), list)
        assert ds.get_userdata_raw() == USERDATA
        assert ds.get_vendordata_raw() == util.load_yaml(METADATA).get(
            "vendor_data"
        )

    @mock.patch("cloudinit.sources.helpers.hetzner.get_metadata")
    @mock.patch("cloudinit.net.find_fallback_nic")
    @mock.patch("cloudinit.sources.DataSourceHetzner.get_hcloud_data")
    def test_not_on_hetzner_returns_false(
        self, m_get_hcloud_data, m_find_fallback, m_get_metadata, ds
    ):
        """If helper 'get_hcloud_data' returns False,
        return False from get_data."""
        m_get_hcloud_data.return_value = (False, None)
        ret = ds.get_data()

        assert not ret
        # These are a white box attempt to ensure it did not search.
        assert 0 == m_find_fallback.call_count
        assert 0 == m_get_metadata.call_count
